# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class Importer(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def translate(self, graph):
        raise NotImplementedError("Correct import strategy must be provided by subclasses.")



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Tuple, List

QM_KEY_NAME = "name"
QM_KEY_VALUE = "value"
QM_KEY_DEFAULT = "default"
QM_KEY_TYPE = "type"

QM_FIELD_NAME = "name"
QM_FIELD_IS_EDITABLE = "is_editable"
QM_FIELD_IS_WRITE = "is_write"
QM_FIELD_PATH = "path"
QM_FIELD_SIGNATURE = "signature"

LEFT_PARANTHESES = "("
RIGHT_PARANTHESES = ")"
EQUALS_DELIMITER = " = "
NAME_TYPE_DELIMITIER = " :: "
COMMA_SEP = ", "
STRING_TYPE = "STRING"
QUOTATION_MARK = '"'


class QueryModule:
    """Class representing a single MAGE query module."""

    def __init__(self, **kwargs) -> None:
        arguments, returns = parse_query_module_signature(kwargs[QM_FIELD_SIGNATURE])

        self.name = kwargs[QM_FIELD_NAME]
        self.is_editable = kwargs[QM_FIELD_IS_EDITABLE]
        self.is_write = kwargs[QM_FIELD_IS_WRITE]
        self.path = kwargs[QM_FIELD_PATH]
        self.signature = kwargs[QM_FIELD_SIGNATURE]
        self.arguments = arguments
        self.returns = returns

    def __str__(self) -> str:
        return self.name

    def set_argument_values(self, **kwargs) -> None:
        """Set values for QueryModule arguments so the module can be called.

        Kwargs:
            Named arguments in self.arguments.

        Raises:
            KeyError: Passed an argument not in the self.arguments list.
        """
        for argument_name in kwargs:
            has_arg = False
            for argument_dict in self.arguments:
                if argument_dict[QM_KEY_NAME] == argument_name:
                    argument_dict[QM_KEY_VALUE] = str(kwargs[argument_name])
                    has_arg = True
                    break
            if not has_arg:
                raise KeyError(f"{argument_name} is not an argument in this query module.")

    def get_arguments_for_call(self) -> str:
        """return inputs in form "value1, value2, ..." for QueryBuilder call()
        method.

        Raises:
            KeyError: Cannot get all values of arguments because one or more is
            not set.
        """
        arguments_str = ""

        for argument_dict in self.arguments:
            if QM_KEY_VALUE in argument_dict:
                val = argument_dict[QM_KEY_VALUE]
            elif QM_KEY_DEFAULT in argument_dict:
                val = argument_dict[QM_KEY_DEFAULT]
            else:
                raise KeyError(f"{argument_dict[QM_KEY_NAME]} has no value set.")

            if argument_dict[QM_KEY_TYPE] == STRING_TYPE:
                arguments_str += QUOTATION_MARK + val + QUOTATION_MARK
            else:
                arguments_str += val

            arguments_str += COMMA_SEP

        return arguments_str[:-2]


def parse_query_module_signature(signature: str) -> Tuple[List[Dict[str, str]], List[Dict[str, str]]]:
    """Query Modules signatures received from Memgraph are parsed into a
    list of dictionaries.

    One list is for arguments and another for returns.
    For instance, if a query module signature is:
    dummy_module.dummy(lst :: LIST OF STRING, num = 3 :: NUMBER) :: (ret :: STRING)
    the method should return a list of arguments:
    [{"name": "lst", "type": "LIST OF STRING"}, {"name": "num", "type": "NUMBER", "default": 3}]
    and a list of returns:
    [{"name": "ret", "type": "STRING"}]

    Dictionary consists of fields: "name" - argument name, "type" - data
    type of argument and "default" where default argument value is given

    Args:
        signature: module signature as returned by Cypher CALL operation
    """
    end_arguments_parentheses = signature.index(RIGHT_PARANTHESES)
    arguments_field = signature[signature.index(LEFT_PARANTHESES) + 1 : end_arguments_parentheses]
    returns_field = signature[
        signature.index(LEFT_PARANTHESES, end_arguments_parentheses)
        + 1 : signature.index(RIGHT_PARANTHESES, end_arguments_parentheses + 1)
    ]

    arguments = parse_field(
        vars_field=arguments_field.strip(),
        name_type_delimiter=NAME_TYPE_DELIMITIER,
        default_value_delimiter=EQUALS_DELIMITER,
    )
    returns = parse_field(
        vars_field=returns_field.strip(),
        name_type_delimiter=NAME_TYPE_DELIMITIER,
        default_value_delimiter=EQUALS_DELIMITER,
    )

    return arguments, returns


def parse_field(
    vars_field: str, name_type_delimiter: str = NAME_TYPE_DELIMITIER, default_value_delimiter: str = EQUALS_DELIMITER
) -> List[Dict[str, str]]:
    """Parse a field of arguments or returns from Query Module signature.

    Args:
        vars_field: signature field inside parentheses
    """
    if len(vars_field) == 0:
        return []

    vars = []

    for var in vars_field.split(COMMA_SEP):
        var_dict = {}
        sides = var.split(name_type_delimiter)
        var_dict[QM_KEY_TYPE] = sides[1]
        if default_value_delimiter in sides[0]:
            splt = sides[0].split(default_value_delimiter)
            var_dict[QM_KEY_NAME] = splt[0]
            var_dict[QM_KEY_DEFAULT] = splt[1].strip(QUOTATION_MARK)
        else:
            var_dict[QM_KEY_NAME] = sides[0]

        vars.append(var_dict)

    return vars



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from pydantic import validator  # noqa F401

from gqlalchemy.models import (  # noqa F401
    MemgraphConstraintExists,
    MemgraphConstraintUnique,
    MemgraphIndex,
    MemgraphKafkaStream,
    MemgraphPulsarStream,
    MemgraphTrigger,
    Neo4jConstraintUnique,
    Neo4jIndex,
    Node,
    Path,
    Relationship,
    Field,
)
from gqlalchemy.disk_storage import SQLitePropertyDatabase  # noqa F401
from gqlalchemy.instance_runner import (  # noqa F401
    DockerImage,
    MemgraphInstanceBinary,
    MemgraphInstanceDocker,
    wait_for_docker_container,
    wait_for_port,
)
from gqlalchemy.exceptions import GQLAlchemyError, GQLAlchemyWarning  # noqa F401

from gqlalchemy.query_builders import (  # noqa F401
    neo4j_query_builder,
    memgraph_query_builder,
    memgraph_query_builder as query_builder,
)
from gqlalchemy.query_builders.declarative_base import (  # noqa F401
    Call,
    Create,
    Foreach,
    InvalidMatchChainException,
    Return,
    Match,
    Merge,
    NoVariablesMatchedException,
    Unwind,
    With,
)
from gqlalchemy.query_builders.memgraph_query_builder import LoadCsv, QueryBuilder
from gqlalchemy.query_builders.neo4j_query_builder import Neo4jQueryBuilder  # noqa F401
from gqlalchemy.vendors.memgraph import Memgraph  # noqa F401
from gqlalchemy.vendors.neo4j import Neo4j  # noqa F401

warnings.filterwarnings("once", category=GQLAlchemyWarning)
__all__ = ["Memgraph"]

call = Call
create = Create
match = Match
merge = Merge
unwind = Unwind
with_ = With
foreach = Foreach
return_ = Return
load_csv = LoadCsv
MemgraphQueryBuilder = QueryBuilder



# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gqlalchemy import Memgraph
from gqlalchemy.transformations.graph_type import GraphType
from gqlalchemy.transformations.importing.importer import Importer

from gqlalchemy.exceptions import raise_if_not_imported
import gqlalchemy.memgraph_constants as mg_consts

try:
    from gqlalchemy.transformations.translators.dgl_translator import DGLTranslator
except ModuleNotFoundError:
    DGLTranslator = None

from gqlalchemy.transformations.translators.nx_translator import NxTranslator

try:
    from gqlalchemy.transformations.translators.pyg_translator import PyGTranslator
except ModuleNotFoundError:
    PyGTranslator = None


class GraphImporter(Importer):
    """Imports dgl, pyg or networkx graph representations to Memgraph.
    The following code will suffice for importing queries.
    >>> importer = GraphImporter("dgl")
    graph = DGLGraph(...)
    importer.translate(graph)  # queries are inserted in this step
    Args:
        graph_type: The type of the graph.
    """

    def __init__(
        self,
        graph_type: str,
        host: str = mg_consts.MG_HOST,
        port: int = mg_consts.MG_PORT,
        username: str = mg_consts.MG_USERNAME,
        password: str = mg_consts.MG_PASSWORD,
        encrypted: bool = mg_consts.MG_ENCRYPTED,
        client_name: str = mg_consts.MG_CLIENT_NAME,
        lazy: bool = mg_consts.MG_LAZY,
    ) -> None:
        super().__init__()
        self.graph_type = graph_type.upper()
        if self.graph_type == GraphType.DGL.name:
            raise_if_not_imported(dependency=DGLTranslator, dependency_name="dgl")
            self.translator = DGLTranslator(host, port, username, password, encrypted, client_name, lazy)
        elif self.graph_type == GraphType.PYG.name:
            raise_if_not_imported(dependency=PyGTranslator, dependency_name="torch_geometric")
            self.translator = PyGTranslator(host, port, username, password, encrypted, client_name, lazy)
        elif self.graph_type == GraphType.NX.name:
            self.translator = NxTranslator(host, port, username, password, encrypted, client_name, lazy)
        else:
            raise ValueError("Unknown import option. Currently supported options are: DGL, PyG and Networkx.")

    def translate(self, graph) -> None:
        """Gets cypher queries using the underlying translator and then inserts all queries to Memgraph DB.
        Args:
            graph: dgl, pytorch geometric or nx graph instance.
        """
        memgraph = Memgraph()
        for query in self.translator.to_cypher_queries(graph):
            memgraph.execute(query)



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from datetime import datetime, date, time, timedelta
from enum import Enum
import inspect
import math
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np

try:
    import torch
except ModuleNotFoundError:
    torch = None

from gqlalchemy.exceptions import raise_if_not_imported


class DatetimeKeywords(Enum):
    DURATION = "duration"
    LOCALTIME = "localTime"
    LOCALDATETIME = "localDateTime"
    DATE = "date"


datetimeKwMapping = {
    timedelta: DatetimeKeywords.DURATION.value,
    time: DatetimeKeywords.LOCALTIME.value,
    datetime: DatetimeKeywords.LOCALDATETIME.value,
    date: DatetimeKeywords.DATE.value,
}


class NanValuesHandle(Enum):
    THROW_EXCEPTION = 1
    REMOVE_PROPERTY = 2


class NetworkXCypherConfig:
    def __init__(self, create_index: bool = False, nan_handler: NanValuesHandle = NanValuesHandle.THROW_EXCEPTION):
        self._create_index = create_index
        self._nan_handler = nan_handler

    @property
    def create_index(self) -> bool:
        return self._create_index

    @property
    def nan_handler(self) -> NanValuesHandle:
        return self._nan_handler


def _format_timedelta(duration: timedelta) -> str:
    days = int(duration.total_seconds() // 86400)
    remainder_sec = duration.total_seconds() - days * 86400
    hours = int(remainder_sec // 3600)
    remainder_sec -= hours * 3600
    minutes = int(remainder_sec // 60)
    remainder_sec -= minutes * 60

    return f"P{days}DT{hours}H{minutes}M{remainder_sec}S"


def _is_torch_tensor(value):
    for cls in inspect.getmro(type(value)):
        try:
            if cls.__module__ == "torch" and cls.__name__ == "Tensor":
                return True
        except Exception:
            pass
    return False


def to_cypher_value(value: Any, config: NetworkXCypherConfig = None) -> str:
    """Converts value to a valid Cypher type."""
    if config is None:
        config = NetworkXCypherConfig()

    value_type = type(value)

    if _is_torch_tensor(value):
        raise_if_not_imported(dependency=torch, dependency_name="torch")

        if value.squeeze().size() == 1:
            return value.squeeze().item()
        else:
            return value.tolist()

    if isinstance(value_type, str) and is_numeric(value):
        return float(value)

    if value_type == CypherVariable:
        return str(value)

    if isinstance(value, (timedelta, time, datetime, date)):
        return f"{datetimeKwMapping[value_type]}('{_format_timedelta(value) if isinstance(value, timedelta) else value.isoformat()}')"

    if value_type == str and value.lower() in ["true", "false", "null"]:
        return value

    if value_type == float and math.isnan(value):
        if config.nan_handler == NanValuesHandle.THROW_EXCEPTION:
            raise NanException("Nan values are not allowed!")

        return "null"

    if value_type in [int, float, bool]:
        return str(value)

    if value_type in [list, set, tuple, np.ndarray]:
        return f"[{', '.join(map(to_cypher_value, value))}]"

    if value_type == dict:
        lines = ", ".join(f"{k}: {to_cypher_value(v)}" for k, v in value.items())
        return f"{{{lines}}}"

    if isinstance(value, np.generic):
        return to_cypher_value(value.item())

    if value is None:
        return "null"

    return f"'{value}'"


def is_numeric(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def to_cypher_properties(properties: Optional[Dict[str, Any]] = None, config=None) -> str:
    """Converts properties to a Cypher key-value properties."""
    if config is None:
        config = NetworkXCypherConfig()

    if not properties:
        return ""

    properties_str = []
    for key, value in properties.items():
        value_str = to_cypher_value(value, config)
        properties_str.append(f"{key}: {value_str}")

    return "{{{}}}".format(", ".join(properties_str))


def to_cypher_labels(labels: Union[str, List[str], None]) -> str:
    """Converts labels to a Cypher label definition."""
    if labels:
        if isinstance(labels, str):
            return f":{labels}"
        return f":{':'.join(labels)}"
    return ""


def to_cypher_qm_arguments(arguments: Optional[Union[str, Tuple[Union[str, int, float]]]]) -> str:
    """Converts query module arguments to a valid Cypher string of query module arguments."""
    if isinstance(arguments, tuple):
        return ", ".join([to_cypher_value(arg) for arg in arguments])

    return arguments


class CypherObject(ABC):
    """Abstract method representing an object in cypher syntax, such as nodes
    and relationships.
    """

    @abstractmethod
    def __str__(self) -> str:
        pass


class CypherNode(CypherObject):
    """Represents a node in Cypher syntax."""

    def __init__(self, variable: str = None, labels: Optional[Union[str, list]] = None) -> None:
        super().__init__()
        if isinstance(labels, str) and labels:
            self._labels = [labels]
        else:
            self._labels = labels
        self._variable = variable

    def __str__(self) -> str:
        s = "("

        if self._variable:
            s += self._variable

        if not self._labels:
            return s + ")"

        s += ":" + ":".join(self._labels) + ")"
        return s


class RelationshipDirection(Enum):
    """Defines the direction of CypherRelationship object."""

    UNDIRECTED = 1
    LEFT = 2
    RIGHT = 3


class CypherRelationship(CypherObject):
    """Represents a relationship in Cypher syntax. Multiple types can not be
    set on a relationship, only queried.
    """

    def __init__(
        self,
        types: Optional[Union[str, list]] = None,
        direction: Optional[RelationshipDirection] = RelationshipDirection.UNDIRECTED,
    ) -> None:
        super().__init__()
        if isinstance(types, str):
            self.types = [types]
        else:
            self.types = types
        self.direction = direction

    def __str__(self) -> str:
        if self.types:
            cypher_relationship = "-[:" + " | :".join(self.types) + "]-"
        else:
            cypher_relationship = "--"

        if self.direction == RelationshipDirection.LEFT:
            return "<" + cypher_relationship
        elif self.direction == RelationshipDirection.RIGHT:
            return cypher_relationship + ">"
        else:
            return cypher_relationship


class CypherVariable:
    """Class for support of using a variable as value in Cypher. Used
    to avoid the quotes given to property values and query module arguments.
    """

    def __init__(self, name: str) -> None:
        self._name = name

    def __str__(self) -> str:
        return self._name


class NanException(Exception):
    pass



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime, date, time, timedelta
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

from pydantic import BaseModel, Extra, Field, PrivateAttr  # noqa F401

from gqlalchemy.exceptions import (
    GQLAlchemyError,
    GQLAlchemySubclassNotFoundWarning,
    GQLAlchemyDatabaseMissingInFieldError,
    GQLAlchemyDatabaseMissingInNodeClassError,
)

# Suppress the warning GQLAlchemySubclassNotFoundWarning
IGNORE_SUBCLASSNOTFOUNDWARNING = False


class DatetimeKeywords(Enum):
    DURATION = "duration"
    LOCALTIME = "localTime"
    LOCALDATETIME = "localDateTime"
    DATE = "date"


datetimeKwMapping = {
    timedelta: DatetimeKeywords.DURATION.value,
    time: DatetimeKeywords.LOCALTIME.value,
    datetime: DatetimeKeywords.LOCALDATETIME.value,
    date: DatetimeKeywords.DATE.value,
}


def _format_timedelta(duration: timedelta) -> str:
    days = int(duration.total_seconds() // 86400)
    remainder_sec = duration.total_seconds() - days * 86400
    hours = int(remainder_sec // 3600)
    remainder_sec -= hours * 3600
    minutes = int(remainder_sec // 60)
    remainder_sec -= minutes * 60

    return f"P{days}DT{hours}H{minutes}M{remainder_sec}S"


class TriggerEventType:
    """An enum representing types of trigger events."""

    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"

    @classmethod
    def list(cls):
        return [cls.CREATE, cls.UPDATE, cls.DELETE]


class TriggerEventObject:
    """An enum representing types of trigger objects.

    NODE -> `()`
    RELATIONSHIP -> `-->`
    """

    NODE = "()"
    RELATIONSHIP = "-->"

    @classmethod
    def list(cls):
        return [cls.NODE, cls.RELATIONSHIP]


class TriggerExecutionPhase:
    """An enum representing types of trigger objects.

    Enum:
        BEFORE
        AFTER
    """

    BEFORE = "BEFORE"
    AFTER = "AFTER"


class FieldAttrsConstants:
    INDEX = "index"
    EXISTS = "exists"
    UNIQUE = "unique"

    @classmethod
    def list(cls):
        return [cls.INDEX, cls.EXISTS, cls.UNIQUE]


@dataclass(frozen=True, eq=True)
class Index(ABC):
    label: str
    property: Optional[str] = None

    def to_cypher(self) -> str:
        return f":{self.label}{f'({self.property})' if self.property else ''}"


@dataclass(frozen=True, eq=True)
class MemgraphIndex(Index):
    pass


@dataclass(frozen=True, eq=True)
class Neo4jIndex(Index):
    type: Optional[str] = None
    uniqueness: Optional[str] = None


@dataclass(frozen=True, eq=True)
class Constraint(ABC):
    label: str

    @abstractmethod
    def to_cypher(self) -> str:
        pass


@dataclass(frozen=True, eq=True)
class MemgraphConstraintUnique(Constraint):
    property: Union[str, Tuple]

    def to_cypher(self) -> str:
        properties_str = ""
        if isinstance(self.property, (tuple, set, list)):
            properties_str = ", ".join([f"n.{prop}" for prop in self.property])
        else:
            properties_str = f"n.{self.property}"
        return f"(n:{self.label}) ASSERT {properties_str} IS UNIQUE"


@dataclass(frozen=True, eq=True)
class MemgraphConstraintExists(Constraint):
    property: str

    def to_cypher(self) -> str:
        return f"(n:{self.label}) ASSERT EXISTS (n.{self.property})"


@dataclass(frozen=True, eq=True)
class Neo4jConstraintUnique(Constraint):
    property: Union[str, Tuple]

    def to_cypher(self) -> str:
        properties_str = ""
        if isinstance(self.property, (tuple, set, list)):
            properties_str = ", ".join([f"n.{prop}" for prop in self.property])
        else:
            properties_str = f"n.{self.property}"
        return f"(n:{self.label}) ASSERT {properties_str} IS UNIQUE"


@dataclass(frozen=True, eq=True)
class Neo4jConstraintExists(Constraint):
    property: str

    def to_cypher(self) -> str:
        return f"(n:{self.label}) ASSERT EXISTS (n.{self.property})"


@dataclass(frozen=True, eq=True)
class MemgraphStream(ABC):
    name: str
    topics: List[str]
    transform: str

    @abstractmethod
    def to_cypher(self) -> str:
        pass


class MemgraphKafkaStream(MemgraphStream):
    """A class for creating and managing Kafka streams in Memgraph.

    Args:
        name: A string representing the stream name.
        topics: A list of strings representing the stream topics.
        transform: A string representing the name of the transformation procedure.
        consumer_group: A string representing the consumer group.
        name: A string representing the batch interval.
        name: A string representing the batch size.
        name: A string or list of strings representing bootstrap server addresses.
    """

    def __init__(
        self,
        name: str,
        topics: List[str],
        transform: str,
        consumer_group: str = None,
        batch_interval: str = None,
        batch_size: str = None,
        bootstrap_servers: Union[str, List[str]] = None,
    ):
        super().__init__(name, topics, transform)
        self.consumer_group = consumer_group
        self.batch_interval = batch_interval
        self.batch_size = batch_size
        self.bootstrap_servers = bootstrap_servers

    def to_cypher(self) -> str:
        """Converts Kafka stream to a Cypher clause."""
        topics = ",".join(self.topics)
        query = f"CREATE KAFKA STREAM {self.name} TOPICS {topics} TRANSFORM {self.transform}"
        if self.consumer_group is not None:
            query += f" CONSUMER_GROUP {self.consumer_group}"
        if self.batch_interval is not None:
            query += f" BATCH_INTERVAL {self.batch_interval}"
        if self.batch_size is not None:
            query += f" BATCH_SIZE {self.batch_size}"
        if self.bootstrap_servers is not None:
            if isinstance(self.bootstrap_servers, str):
                servers_field = f"'{self.bootstrap_servers}'"
            else:
                servers_field = str(self.bootstrap_servers)[1:-1]
            query += f" BOOTSTRAP_SERVERS {servers_field}"
        query += ";"
        return query


class MemgraphPulsarStream(MemgraphStream):
    """A class for creating and managing Pulsar streams in Memgraph.

    Args:
        name: A string representing the stream name.
        topics: A list of strings representing the stream topics.
        transform: A string representing the name of the transformation procedure.
        consumer_group: A string representing the consumer group.
        name: A string representing the batch interval.
        name: A string representing the batch size.
        name: A string or list of strings representing bootstrap server addresses.
    """

    def __init__(
        self,
        name: str,
        topics: List[str],
        transform: str,
        batch_interval: str = None,
        batch_size: str = None,
        service_url: str = None,
    ):
        super().__init__(name, topics, transform)
        self.batch_interval = batch_interval
        self.batch_size = batch_size
        self.service_url = service_url

    def to_cypher(self) -> str:
        """Converts Pulsar stream to a Cypher clause."""
        topics = ",".join(self.topics)
        query = f"CREATE PULSAR STREAM {self.name} TOPICS {topics} TRANSFORM {self.transform}"
        if self.batch_interval is not None:
            query += f" BATCH_INTERVAL {self.batch_interval}"
        if self.batch_size is not None:
            query += f" BATCH_SIZE {self.batch_size}"
        if self.service_url is not None:
            query += f" SERVICE_URL {self.service_url}"
        query += ";"
        return query


@dataclass(frozen=True, eq=True)
class MemgraphTrigger:
    name: str
    execution_phase: TriggerExecutionPhase
    statement: str
    event_type: Optional[TriggerEventType] = None
    event_object: Optional[TriggerEventObject] = None

    def to_cypher(self) -> str:
        """Converts a Trigger to a cypher clause."""
        query = f"CREATE TRIGGER {self.name} "
        if self.event_type in TriggerEventType.list():
            query += f"ON " + (
                f"{self.event_object} {self.event_type} "
                if self.event_object in TriggerEventObject.list()
                else f"{self.event_type} "
            )
        query += f"{self.execution_phase} COMMIT EXECUTE "
        query += f"{self.statement};"
        return query


class GraphObject(BaseModel):
    _subtypes_ = dict()

    class Config:
        extra = Extra.allow

    def __init_subclass__(cls, type=None, label=None, labels=None, index=None, db=None):
        """Stores the subclass by type if type is specified, or by class name
        when instantiating a subclass.
        """
        if type is not None:  # Relationship
            cls._subtypes_[type] = cls
        elif label is not None:  # Node
            cls._subtypes_[label] = cls
        else:
            cls._subtypes_[cls.__name__] = cls

    @classmethod
    def __get_validators__(cls):
        yield cls._convert_to_real_type_

    @classmethod
    def _convert_to_real_type_(cls, data):
        """Converts the GraphObject class into the appropriate subclass.
        This is used when deserialising a json representation of the class,
        or the object returned from the GraphDatabase.
        """
        sub = None
        if "_type" in data:  # Relationship
            sub = cls._subtypes_.get(data.get("_type"))

        if "_labels" in data:  # Node
            # Find class that has the most super classes
            labels = data["_labels"]
            classes = [cls._subtypes_[label] for label in labels if label in cls._subtypes_]
            counter = defaultdict(int)
            for class1 in classes:
                counter[class1] += sum(issubclass(class1, class2) for class2 in classes)
            if counter:
                sub = max(counter, key=counter.get)

        if sub is None:
            types = data.get("_type", data.get("_labels"))
            if not IGNORE_SUBCLASSNOTFOUNDWARNING:
                warnings.warn(GQLAlchemySubclassNotFoundWarning(types, cls))

            sub = cls

        return sub(**data)

    @classmethod
    def parse_obj(cls, obj):
        """Used to convert a dictionary object into the appropriate
        GraphObject.
        """
        return cls._convert_to_real_type_(obj)

    def escape_value(
        self, value: Union[None, bool, int, float, str, list, dict, datetime, timedelta, date, time]
    ) -> str:
        value_type = type(value)

        if value is None:
            "Null"
        elif value_type == bool:
            return repr(value)
        elif value_type == int:
            return repr(value)
        elif value_type == float:
            return repr(value)
        elif isinstance(value, str):
            return repr(value) if value.isprintable() else rf"'{value}'"
        elif isinstance(value, list):
            return "[" + ", ".join(self.escape_value(val) for val in value) + "]"
        elif value_type == dict:
            return "{" + ", ".join(f"{key}: {self.escape_value(val)}" for key, val in value.items()) + "}"
        if isinstance(value, (timedelta, time, datetime, date)):
            return f"{datetimeKwMapping[value_type]}('{_format_timedelta(value) if isinstance(value, timedelta) else value.isoformat()}')"
        else:
            raise GQLAlchemyError(
                f"Unsupported value data type: {type(value)}."
                + " Memgraph supports the following data types:"
                + " None, bool, int, float, str, list, dict, datetime."
            )

    def _get_cypher_field_assignment_block(self, variable_name: str, operator: str) -> str:
        """Creates a cypher field assignment block joined using the `operator`
        argument.
        Example:
            self = {"name": "John", "age": 34}
            variable_name = "user"
            operator = " AND "

            returns:
                "user.name = 'John' AND user.age = 34"
        """
        cypher_fields = []
        for field in self.__fields__:
            value = getattr(self, field)
            if value is not None:
                cypher_fields.append(f"{variable_name}.{field} = {self.escape_value(value)}")

        return " " + operator.join(cypher_fields) + " "

    def _get_cypher_fields_or_block(self, variable_name: str) -> str:
        """Returns a cypher field assignment block separated by an OR
        statement.
        """
        return self._get_cypher_field_assignment_block(variable_name, " OR ")

    def _get_cypher_fields_and_block(self, variable_name: str) -> str:
        """Returns a cypher field assignment block separated by an AND
        statement.
        """
        return self._get_cypher_field_assignment_block(variable_name, " AND ")

    def _get_cypher_fields_xor_block(self, variable_name: str) -> str:
        """Returns a cypher field assignment block separated by an XOR
        statement.
        """
        return self._get_cypher_field_assignment_block(variable_name, " XOR ")

    # TODO: add NOT

    def _get_cypher_set_properties(self, variable_name: str) -> str:
        """Returns a cypher set properties block."""
        cypher_set_properties = []
        for field in self.__fields__:
            attributes = self.__fields__[field].field_info.extra
            value = getattr(self, field)
            if value is not None and not attributes.get("on_disk", False):
                cypher_set_properties.append(f" SET {variable_name}.{field} = {self.escape_value(value)}")

        return " " + " ".join(cypher_set_properties) + " "

    def __str__(self) -> str:
        return "<GraphObject>"

    def __repr__(self) -> str:
        return str(self)


class UniqueGraphObject(GraphObject):
    _id: Optional[Any] = PrivateAttr()
    _properties: Optional[Dict[str, Any]] = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._id = data.get("_id")
        self._type = data.get("_type")

    @property
    def _properties(self) -> Dict[str, Any]:
        return {k: v for k, v in dict(self).items() if not k.startswith("_") and k != "labels"}

    def __str__(self) -> str:
        return f"<GraphObject id={self._id} properties={self._properties}>"

    def __repr__(self) -> str:
        return str(self)


class NodeMetaclass(BaseModel.__class__):
    def __new__(mcs, name, bases, namespace, **kwargs):  # noqa C901
        """This creates the class `Node`. It also creates all subclasses
        of `Node`. Whenever a class is defined as a subclass of `Node`,
        `MyMeta.__new__` is called.
        """

        def field_in_superclass(field, constraint):
            nonlocal bases
            for base in bases:
                if field in base.__fields__:
                    attrs = base.__fields__[field].field_info.extra
                    if constraint in attrs:
                        return base

            return None

        def get_base_labels() -> Set[str]:
            base_labels = set()
            nonlocal bases
            for base in bases:
                if hasattr(base, "labels"):
                    base_labels = base_labels.union(base.labels)

            return base_labels

        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        cls.index = kwargs.get("index")
        cls.label = kwargs.get("label", name)
        if name != "Node":
            cls.labels = get_base_labels().union({cls.label}, kwargs.get("labels", set()))

        db = kwargs.get("db")
        if cls.index is True:
            if db is None:
                raise GQLAlchemyDatabaseMissingInNodeClassError(cls=cls)

            index = MemgraphIndex(cls.label)
            db.create_index(index)

        for field in cls.__fields__:
            attrs = cls.__fields__[field].field_info.extra
            field_type = cls.__fields__[field].type_.__name__
            label = attrs.get("label", cls.label)
            skip_constraints = False

            if db is None:
                db = attrs.get("db")

            for constraint in FieldAttrsConstants.list():
                if constraint in attrs and db is None:
                    base = field_in_superclass(field, constraint)
                    if base is not None:
                        cls.__fields__[field].field_info.extra = base.__fields__[field].field_info.extra
                        skip_constraints = True
                        break

                    raise GQLAlchemyDatabaseMissingInFieldError(
                        constraint=constraint,
                        field=field,
                        field_type=field_type,
                    )

            if skip_constraints:
                continue

            if FieldAttrsConstants.INDEX in attrs and attrs[FieldAttrsConstants.INDEX] is True:
                index = MemgraphIndex(label, field)
                db.create_index(index)

            if FieldAttrsConstants.EXISTS in attrs and attrs[FieldAttrsConstants.EXISTS] is True:
                constraint = MemgraphConstraintExists(label, field)
                db.create_constraint(constraint)

            if FieldAttrsConstants.UNIQUE in attrs and attrs[FieldAttrsConstants.UNIQUE] is True:
                constraint = MemgraphConstraintUnique(label, field)
                db.create_constraint(constraint)

            if attrs and "db" in attrs:
                del attrs["db"]

        return cls


class Node(UniqueGraphObject, metaclass=NodeMetaclass):
    _labels: Set[str] = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._labels = data.get("_labels", getattr(type(self), "labels", {"Node"}))

    def __str__(self) -> str:
        return "".join(
            (
                f"<{type(self).__name__}",
                f" id={self._id}",
                f" labels={self._labels}",
                f" properties={self._properties}",
                ">",
            )
        )

    def _get_cypher_unique_fields_or_block(self, variable_name: str) -> str:
        """Get's a cypher assignment block using the unique fields."""
        cypher_unique_fields = []
        for field in self.__fields__:
            attrs = self.__fields__[field].field_info.extra
            if "unique" in attrs:
                value = getattr(self, field)
                if value is not None:
                    cypher_unique_fields.append(f"{variable_name}.{field} = {self.escape_value(value)}")

        return " " + " OR ".join(cypher_unique_fields) + " "

    def has_unique_fields(self) -> bool:
        """Returns True if the Node has any unique fields."""
        for field in self.__fields__:
            if "unique" in self.__fields__[field].field_info.extra:
                if getattr(self, field) is not None:
                    return True
        return False

    @property
    def _label(self) -> str:
        return ":".join(sorted(self._labels))

    def save(self, db: "Database") -> "Node":  # noqa F821
        """Saves node to Memgraph.
        If the node._id is not None it fetches the node with the same id from
        Memgraph and updates it's fields.
        If the node has unique fields it fetches the nodes with the same unique
        fields from Memgraph and updates it's fields.
        Otherwise it creates a new node with the same properties.
        Null properties are ignored.
        """
        node = db.save_node(self)
        for field in self.__fields__:
            setattr(self, field, getattr(node, field))
        self._id = node._id
        return self

    def load(self, db: "Database") -> "Node":  # noqa F821
        """Loads a node from Memgraph.
        If the node._id is not None it fetches the node from Memgraph with that
        internal id.
        If the node has unique fields it fetches the node from Memgraph with
        those unique fields set.
        Otherwise it tries to find any node in Memgraph that has all properties
        set to exactly the same values.
        If no node is found or no properties are set it raises a GQLAlchemyError.
        """
        node = db.load_node(self)
        for field in self.__fields__:
            setattr(self, field, getattr(node, field))
        self._id = node._id
        return self


class RelationshipMetaclass(BaseModel.__class__):
    def __new__(mcs, name, bases, namespace, **kwargs):  # noqa C901
        """This creates the class `Relationship`. It also creates all
        subclasses of `Relationship`. Whenever a class is defined as a
        subclass of `Relationship`, `self.__new__` is called.
        """
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        if name != "Relationship":
            cls.type = kwargs.get("type", name)

        return cls


class Relationship(UniqueGraphObject, metaclass=RelationshipMetaclass):
    _end_node_id: int = PrivateAttr()
    _start_node_id: int = PrivateAttr()
    _type: str = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._start_node_id = data.get("_start_node_id")
        self._end_node_id = data.get("_end_node_id")
        self._type = data.get("_type", getattr(type(self), "type", "Relationship"))

    @property
    def _nodes(self) -> Tuple[int, int]:
        return (self._start_node_id, self._end_node_id)

    def __str__(self) -> str:
        return "".join(
            (
                f"<{type(self).__name__}",
                f" id={self._id}",
                f" start_node_id={self._start_node_id}",
                f" end_node_id={self._end_node_id}",
                f" nodes={self._nodes}",
                f" type={self._type}",
                f" properties={self._properties}",
                ">",
            )
        )

    def save(self, db: "Database") -> "Relationship":  # noqa F821
        """Saves a relationship to Memgraph.
        If relationship._id is not None it finds the relationship in Memgraph
        and updates it's properties with the values in `relationship`.
        If relationship._id is None, it creates a new relationship.
        If you want to set a relationship._id instead of creating a new
        relationship, use `load_relationship` first.
        """
        relationship = db.save_relationship(self)
        for field in self.__fields__:
            setattr(self, field, getattr(relationship, field))
        self._id = relationship._id
        return self

    def load(self, db: "Database") -> "Relationship":  # noqa F821
        """Returns a relationship loaded from Memgraph.
        If the relationship._id is not None it fetches the relationship from
        Memgraph that has the same internal id.
        Otherwise it returns the relationship whose relationship._start_node_id
        and relationship._end_node_id and all relationship properties that
        are not None match the relationship in Memgraph.
        If there is no relationship like that in Memgraph, or if there are
        multiple relationships like that in Memgraph, throws GQLAlchemyError.
        """
        relationship = db.load_relationship(self)
        for field in self.__fields__:
            setattr(self, field, getattr(relationship, field))
        self._id = relationship._id
        return self


class Path(GraphObject):
    _nodes: Iterable[Node] = PrivateAttr()
    _relationships: Iterable[Relationship] = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._nodes = data.get("_nodes")
        self._relationships = data.get("_relationships")

    def __str__(self) -> str:
        return "".join(
            (
                f"<{type(self).__name__}",
                f" nodes={self._nodes}",
                f" relationships={self._relationships}" ">",
            )
        )



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Union

from gqlalchemy.query_builders.declarative_base import DeclarativeBase
from gqlalchemy.query_builders.memgraph_query_builder import QueryBuilder
from gqlalchemy.vendors.memgraph import Connection, Memgraph


class MemgraphQueryBuilder(QueryBuilder):
    """
    This query builder extends the usual Cypher query builder capabilities with Memgraph's query modules.
    User gets with this module autocomplete features of graph algorithms.
    Documentation on the methods can be found on Memgraph's web page.
    """

    def __init__(self, connection: Optional[Union[Connection, Memgraph]] = None):
        super().__init__(connection)

    def example_c_procedure(self, required_arg: Any, optional_arg=None) -> DeclarativeBase:
        return self.call("example.procedure", (required_arg, optional_arg))

    def example_c_write_procedure(self, required_arg: str) -> DeclarativeBase:
        return self.call("example.write_procedure", (required_arg))

    def graph_analyzer_analyze(self, analyses: Optional[List[str]] = None) -> DeclarativeBase:
        return self.call("graph_analyzer.analyze", (analyses))

    def graph_analyzer_analyze_subgraph(
        self, vertices: List[Any], edges: List[Any], analyses: Optional[List[str]] = None
    ) -> DeclarativeBase:
        return self.call("graph_analyzer.analyze_subgraph", (vertices, edges, analyses))

    def graph_analyzer_help(self, name: str, value: str) -> DeclarativeBase:
        return self.call("graph_analyzer.help", (name, value))

    def mg_create_module_file(self, filename: str, content: str) -> DeclarativeBase:
        return self.call("mg.create_module_file", (filename, content))

    def mg_delete_module_file(self, path: str) -> DeclarativeBase:
        return self.call("mg.delete_module_file", (path))

    def mg_functions(self) -> DeclarativeBase:
        return self.call("mg.functions")

    def mg_get_module_file(self, path: str) -> DeclarativeBase:
        return self.call("mg.get_module_file", (path))

    def mg_get_module_files(self, is_editable: bool, path: str) -> DeclarativeBase:
        return self.call("mg.get_module_files", (is_editable, path))

    def mg_kafka_set_stream_offset(self, stream_name: str, offset: int) -> DeclarativeBase:
        return self.call("mg.kafka_set_stream_offset", (stream_name, offset))

    def mg_kafka_stream_info(self, stream_name: str) -> DeclarativeBase:
        return self.call("mg.kafka_stream_info", (stream_name))

    def mg_load(self, module_name: str) -> DeclarativeBase:
        return self.call("mg.load", (module_name))

    def mg_load_all(self) -> DeclarativeBase:
        return self.call("mg.load_all")

    def mg_procedures(self) -> DeclarativeBase:
        return self.call("mg.procedures")

    def mg_pulsar_stream_info(self, stream_name: str) -> DeclarativeBase:
        return self.call("mg.procedures", (stream_name))

    def mg_transformations(self) -> DeclarativeBase:
        return self.call("mg.transformations")

    def mg_update_module_file(self, path: str, content: str) -> DeclarativeBase:
        return self.call("mg.update_module_file", (path, content))

    def nxalg_all_shortest_paths(
        self, source: Any, target: Any, weight: Optional[str] = None, method: str = "dijkstra"
    ) -> DeclarativeBase:
        return self.call("nxalg.all_shortest_paths", (source, target, weight, method))

    def nxalg_all_simple_paths(self, source: Any, target: Any, cutoff: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.all_simple_paths", (source, target, cutoff))

    def nxalg_ancestors(self, source: Any) -> DeclarativeBase:
        return self.call("nxalg.all_simple_paths", (source))

    def nxalg_betweenness_centrality(
        self,
        k: Optional[int] = None,
        normalized: bool = True,
        weight: Optional[str] = None,
        endpoints: bool = False,
        seed: Optional[int] = None,
    ) -> DeclarativeBase:
        return self.call("nxalg.betweenness_centrality", (k, normalized, weight, endpoints, seed))

    def nxalg_bfs_edges(self, source: Any, reverse: bool = False, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.bfs_edges", (source, reverse, depth_limit))

    def nxalg_bfs_predecessors(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.bfs_predecessors", (source, depth_limit))

    def nxalg_bfs_successors(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.bfs_successors", (source, depth_limit))

    def nxalg_bfs_tree(self, source: Any, reverse: bool = False, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.bfs_tree", (source, reverse, depth_limit))

    def nxalg_biconnected_components(self) -> DeclarativeBase:
        return self.call("nxalg.biconnected_components")

    def nxalg_bridges(self, root: Any) -> DeclarativeBase:
        return self.call("nxalg.bridges", (root))

    def nxalg_center(self) -> DeclarativeBase:
        return self.call("nxalg.center")

    def nxalg_chain_decomposition(self, root: Any) -> DeclarativeBase:
        return self.call("nxalg.chain_decomposition", (root))

    def nxalg_check_planarity(self) -> DeclarativeBase:
        return self.call("nxalg.check_planarity")

    def nxalg_clustering(self, nodes: Optional[List[Any]] = None, weight: Optional[str] = None) -> DeclarativeBase:
        return self.call("nxalg.clustering", (nodes, weight))

    def nxalg_communicability(self) -> DeclarativeBase:
        return self.call("nxalg.communicability")

    def nxalg_core_number(self) -> DeclarativeBase:
        return self.call("nxalg.core_number")

    def nxalg_degree_assortativity_coefficient(
        self, x: str = "out", y: str = "in", weight: Optional[str] = None, nodes: Optional[List[Any]] = None
    ) -> DeclarativeBase:
        return self.call("nxalg.degree_assortativity_coefficient", (x, y, weight, nodes))

    def nxalg_descendants(self, source: Any) -> DeclarativeBase:
        return self.call("nxalg.descendants", (source))

    def nxalg_dfs_postorder_nodes(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.dfs_postorder_nodes", (source, depth_limit))

    def nxalg_dfs_predecessors(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.dfs_predecessors", (source, depth_limit))

    def nxalg_dfs_preorder_nodes(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.dfs_preorder_nodes", (source, depth_limit))

    def nxalg_dfs_successors(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.dfs_successors", (source, depth_limit))

    def nxalg_dfs_tree(self, source: Any, depth_limit: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.dfs_tree", (source, depth_limit))

    def nxalg_diameter(self) -> DeclarativeBase:
        return self.call("nxalg.diameter")

    def nxalg_dominance_frontiers(self, start: Any) -> DeclarativeBase:
        return self.call("nxalg.dominance_frontiers", (start))

    def nxalg_dominating_set(self, start: Any) -> DeclarativeBase:
        return self.call("nxalg.dominance_frontiers", (start))

    def nxalg_edge_bfs(self, source: Optional[Any], orientation: Optional[str] = None) -> DeclarativeBase:
        return self.call("nxalg.edge_bfs", (source, orientation))

    def nxalg_edge_dfs(self, source: Optional[Any], orientation: Optional[str] = None) -> DeclarativeBase:
        return self.call("nxalg.edge_dfs", (source, orientation))

    def nxalg_find_cliques(self) -> DeclarativeBase:
        return self.call("nxalg.find_cliques")

    def nxalg_find_cycle(
        self, source: Optional[List[Any]] = None, orientation: Optional[str] = None
    ) -> DeclarativeBase:
        return self.call("nxalg.find_cycle", (source, orientation))

    def nxalg_flow_hierarchy(self, weight: Optional[str] = None) -> DeclarativeBase:
        return self.call("nxalg.flow_hierarchy", (weight))

    def nxalg_global_efficiency(self) -> DeclarativeBase:
        return self.call("nxalg.global_efficiency")

    def nxalg_greedy_color(self, strategy: str = "largest_first", interchange: bool = False) -> DeclarativeBase:
        return self.call("nxalg.greedy_color", (strategy, interchange))

    def nxalg_has_eulerian_path(self) -> DeclarativeBase:
        return self.call("nxalg.has_eulerian_path")

    def nxalg_has_path(self, source: Any, target: Any) -> DeclarativeBase:
        return self.call("nxalg.has_path", (source, target))

    def nxalg_immediate_dominators(self, start: Any) -> DeclarativeBase:
        return self.call("nxalg.immediate_dominators", (start))

    def nxalg_is_arborescence(self) -> DeclarativeBase:
        return self.call("nxalg.is_arborescence")

    def nxalg_is_at_free(self) -> DeclarativeBase:
        return self.call("nxalg.is_at_free")

    def nxalg_is_bipartite(self) -> DeclarativeBase:
        return self.call("nxalg.is_bipartite")

    def nxalg_is_branching(self) -> DeclarativeBase:
        return self.call("nxalg.is_branching")

    def nxalg_is_chordal(self) -> DeclarativeBase:
        return self.call("nxalg.is_chordal")

    def nxalg_is_distance_regular(self) -> DeclarativeBase:
        return self.call("nxalg.is_distance_regular")

    def nxalg_is_edge_cover(self, cover: List[Any]) -> DeclarativeBase:
        return self.call("nxalg.is_edge_cover", (cover))

    def nxalg_is_eulerian(self) -> DeclarativeBase:
        return self.call("nxalg.is_eulerian")

    def nxalg_is_forest(self) -> DeclarativeBase:
        return self.call("nxalg.is_forest")

    def nxalg_is_isolate(self, n: Any) -> DeclarativeBase:
        return self.call("nxalg.is_isolate", (n))

    def nxalg_is_isomorphic(
        self, nodes1: List[Any], edges1: List[Any], nodes2: List[Any], edges2: List[Any]
    ) -> DeclarativeBase:
        return self.call("nxalg.is_isomorphic", (nodes1, edges1, nodes2, edges2))

    def nxalg_is_semieulerian(self) -> DeclarativeBase:
        return self.call("nxalg.is_semieulerian")

    def nxalg_is_simple_path(self, nodes: List[Any]) -> DeclarativeBase:
        return self.call("nxalg.is_simple_path", (nodes))

    def nxalg_is_strongly_regular(self) -> DeclarativeBase:
        return self.call("nxalg.is_strongly_regular")

    def nxalg_is_tournament(self) -> DeclarativeBase:
        return self.call("nxalg.is_tournament")

    def nxalg_is_tree(self) -> DeclarativeBase:
        return self.call("nxalg.is_tree")

    def nxalg_isolates(self) -> DeclarativeBase:
        return self.call("nxalg.isolates")

    def nxalg_jaccard_coefficient(self, ebunch: Optional[List[Any]] = None) -> DeclarativeBase:
        return self.call("nxalg.jaccard_coefficient", (ebunch))

    def nxalg_k_clique_communities(self, k: int, cliques: List[List[Any]] = None) -> DeclarativeBase:
        return self.call("nxalg.k_clique_communities", (k, cliques))

    def nxalg_k_components(self, density: float = 0.95) -> DeclarativeBase:
        return self.call("nxalg.k_components", (density))

    def nxalg_k_edge_components(self, k: int) -> DeclarativeBase:
        return self.call("nxalg.k_edge_components", (k))

    def nxalg_local_efficiency(self) -> DeclarativeBase:
        return self.call("nxalg.local_efficiency")

    def nxalg_lowest_common_ancestor(self, node1: Any, node2: Any) -> DeclarativeBase:
        return self.call("nxalg.lowest_common_ancestor", (node1, node2))

    def nxalg_maximal_matching(self) -> DeclarativeBase:
        return self.call("nxalg.maximal_matching")

    def nxalg_minimum_spanning_tree(
        self, weight: str = "weight", algorithm: str = "kruskal", ignore_nan: bool = False
    ) -> DeclarativeBase:
        return self.call("nxalg.minimum_spanning_tree", (weight, algorithm, ignore_nan))

    def nxalg_multi_source_dijkstra_path(
        self, sources: List[Any], cutoff: Optional[int] = None, weight: str = "weight"
    ) -> DeclarativeBase:
        return self.call("nxalg.multi_source_dijkstra_path", (sources, cutoff, weight))

    def nxalg_multi_source_dijkstra_path_length(
        self, sources: List[Any], cutoff: Optional[int] = None, weight: str = "weight"
    ) -> DeclarativeBase:
        return self.call("nxalg.multi_source_dijkstra_path_length", (sources, cutoff, weight))

    def nxalg_node_boundary(self, nbunch1: List[Any], nbunch2: List[Any] = None) -> DeclarativeBase:
        return self.call("nxalg.node_boundary", (nbunch1, nbunch2))

    def nxalg_node_connectivity(self, source: List[Any] = None, target: List[Any] = None) -> DeclarativeBase:
        return self.call("nxalg.node_connectivity", (source, target))

    def nxalg_node_expansion(self, s: List[Any]) -> DeclarativeBase:
        return self.call("nxalg.node_expansion", (s))

    def nxalg_non_randomness(self, k: Optional[int] = None) -> DeclarativeBase:
        return self.call("nxalg.non_randomness", (k))

    def nxalg_pagerank(
        self,
        alpha: float = 0.85,
        personalization: Optional[str] = None,
        max_iter: int = 100,
        tol: float = 1e-06,
        nstart: Optional[str] = None,
        weight: Optional[str] = "weight",
        dangling: Optional[str] = None,
    ) -> DeclarativeBase:
        return self.call("nxalg.pagerank", (alpha, personalization, max_iter, tol, nstart, weight, dangling))

    def nxalg_reciprocity(self, nodes: List[Any] = None) -> DeclarativeBase:
        return self.call("nxalg.reciprocity", (nodes))

    def nxalg_shortest_path(
        self,
        source: Optional[Any] = None,
        target: Optional[Any] = None,
        weight: Optional[str] = None,
        method: str = "dijkstra",
    ) -> DeclarativeBase:
        return self.call("nxalg.shortest_path", (source, target, weight, method))

    def nxalg_shortest_path_length(
        self,
        source: Optional[Any] = None,
        target: Optional[Any] = None,
        weight: Optional[str] = None,
        method: str = "dijkstra",
    ) -> DeclarativeBase:
        return self.call("nxalg.shortest_path_length", (source, target, weight, method))

    def nxalg_simple_cycles(self) -> DeclarativeBase:
        return self.call("nxalg.simple_cycles")

    def nxalg_strongly_connected_components(self) -> DeclarativeBase:
        return self.call("nxalg.strongly_connected_components")

    def nxalg_topological_sort(self) -> DeclarativeBase:
        return self.call("nxalg.topological_sort")

    def nxalg_triadic_census(self) -> DeclarativeBase:
        return self.call("nxalg.triadic_census")

    def nxalg_voronoi_cells(self, center_nodes: List[Any], weight: str = "weight") -> DeclarativeBase:
        return self.call("nxalg.voronoi_cells", (center_nodes, weight))

    def nxalg_wiener_index(self, weight: Optional[str] = None) -> DeclarativeBase:
        return self.call("nxalg.wiener_index", (weight))

    def py_example_procedure(self, required_arg: Optional[Any], optional_arg: Optional[Any] = None) -> DeclarativeBase:
        return self.call("py_example.procedure", (required_arg, optional_arg))

    def py_example_write_procedure(self, property_name: str, property_value: Optional[Any]) -> DeclarativeBase:
        return self.call("py_example.write_procedure", (property_name, property_value))

    def wcc_get_components(self, vertices: List[Any], edges: List[Any]) -> DeclarativeBase:
        return self.call("wcc.get_components", (vertices, edges))


class MageQueryBuilder(MemgraphQueryBuilder):
    """
    This query builder extends the Memgraph query builder with Memgraph MAGE graph algorithm Cypher options.
    User gets with this module autocomplete features of graph algorithms written in MAGE library.
    Documentation on the methods can be found on Memgraph's web page.
    """

    def __init__(self, connection: Optional[Union[Connection, Memgraph]] = None):
        super().__init__(connection)

    def betweenness_centrality_get(
        self, directed: bool = True, normalized: bool = True, threads: int = 8
    ) -> DeclarativeBase:
        return self.call("betweenness_centrality.get", (directed, normalized, threads))

    def betweenness_centrality_online_get(self, normalize: bool = True) -> DeclarativeBase:
        return self.call("betweenness_centrality_online.get", (normalize))

    def betweenness_centrality_online_reset(self) -> DeclarativeBase:
        return self.call("betweenness_centrality_online.reset")

    def betweenness_centrality_online_set(self, normalize: bool = True, threads: int = 8) -> DeclarativeBase:
        return self.call("betweenness_centrality_online.set", (normalize, threads))

    def betweenness_centrality_online_update(
        self,
        created_vertices: List[Any] = [],
        created_edges: List[Any] = [],
        deleted_vertices: List[Any] = [],
        deleted_edges: List[Any] = [],
        normalize: bool = True,
        threads: int = 8,
    ) -> DeclarativeBase:
        return self.call(
            "betweenness_centrality_online.update",
            (created_vertices, created_edges, deleted_vertices, deleted_edges, normalize, threads),
        )

    def biconnected_components_get(self) -> DeclarativeBase:
        return self.call("biconnected_components.get")

    def bipartite_matching_max(self) -> DeclarativeBase:
        return self.call("bipartite_matching.max")

    def bridges_get(self) -> DeclarativeBase:
        return self.call("bridges.get")

    def community_detection_get(
        self,
        weight: str = "weight",
        coloring: bool = False,
        min_graph_shrink: int = 100000,
        community_alg_threshold: float = 1e-06,
        coloring_alg_threshold: float = 0.01,
    ) -> DeclarativeBase:
        return self.call(
            "community_detection.get",
            (weight, coloring, min_graph_shrink, community_alg_threshold, coloring_alg_threshold),
        )

    def community_detection_online_get(self) -> DeclarativeBase:
        return self.call("community_detection_online.get")

    def community_detection_online_reset(self) -> DeclarativeBase:
        return self.call("community_detection_online.reset")

    def community_detection_online_set(
        self,
        directed: bool = False,
        weighted: bool = False,
        similarity_threshold: float = 0.7,
        exponent: float = 4,
        min_value: float = 0.1,
        weight_property: str = "weight",
        w_selfloop: float = 1.0,
        max_iterations: int = 100,
        max_updates: int = 5,
    ) -> DeclarativeBase:
        return self.call(
            "community_detection_online.set",
            (
                directed,
                weighted,
                similarity_threshold,
                exponent,
                min_value,
                weight_property,
                w_selfloop,
                max_iterations,
                max_updates,
            ),
        )

    def community_detection_online_update(
        self,
        createdVertices: List[Any] = [],
        createdEdges: List[Any] = [],
        updatedVertices: List[Any] = [],
        updatedEdges: List[Any] = [],
        deletedVertices: List[Any] = [],
        deletedEdges: List[Any] = [],
    ) -> DeclarativeBase:
        return self.call(
            "community_detection_online.update",
            (createdVertices, createdEdges, updatedVertices, updatedEdges, deletedVertices, deletedEdges),
        )

    def cycles_get(self) -> DeclarativeBase:
        return self.call("cycles.get")

    def distance_calculator_multiple(
        self, start_points: List[Any], end_points: List[Any], metrics: str = "m"
    ) -> DeclarativeBase:
        return self.call("distance_calculator.multiple", (start_points, end_points, metrics))

    def distance_calculator_single(
        self, start: Optional[Any], end: Optional[Any], metrics: str = "m"
    ) -> DeclarativeBase:
        return self.call("distance_calculator.single", (start, end, metrics))

    def export_util_json(self, path: str) -> DeclarativeBase:
        return self.call("export_util.json", (path))

    def graph_coloring_color_graph(
        self, parameters: Dict[str, Union[str, int]], edge_property: str = "weight"
    ) -> DeclarativeBase:
        return self.call("graph_coloring.color_graph", (parameters, edge_property))

    def graph_coloring_color_subgraph(
        self,
        vertices: List[Any],
        edges: List[Any],
        parameters: Dict[str, Union[str, int]],
        edge_property: str = "weight",
    ) -> DeclarativeBase:
        return self.call("graph_coloring.color_subgraph", (vertices, edges, parameters, edge_property))

    def import_util_json(self, path: str) -> DeclarativeBase:
        return self.call("import_util.json", (path))

    def json_util_load_from_path(self, path: str) -> DeclarativeBase:
        return self.call("json_util.load_from_path", (path))

    def json_util_load_from_url(self, url: str) -> DeclarativeBase:
        return self.call("json_util.load_from_url", (url))

    def katz_centrality_get(self, alpha: float = 0.2, epsilon: float = 0.01) -> DeclarativeBase:
        return self.call("katz_centrality.get", (alpha, epsilon))

    def katz_centrality_online_get(self) -> DeclarativeBase:
        return self.call("katz_centrality_online.get")

    def katz_centrality_online_reset(self) -> DeclarativeBase:
        return self.call("katz_centrality_online.reset")

    def katz_centrality_online_set(self, alpha: float = 0.2, epsilon: float = 0.01) -> DeclarativeBase:
        return self.call("katz_centrality_online.set", (alpha, epsilon))

    def katz_centrality_online_update(
        self,
        created_vertices: List[Any] = [],
        created_edges: List[Any] = [],
        deleted_vertices: List[Any] = [],
        deleted_edges: List[Any] = [],
    ) -> DeclarativeBase:
        return self.call(
            "katz_centrality_online.update", (created_vertices, created_edges, deleted_vertices, deleted_edges)
        )

    def max_flow_get_flow(self, start_v: Any, end_v: Any, edge_property: str = "weight") -> DeclarativeBase:
        return self.call("max_flow.get_flow", (start_v, end_v, edge_property))

    def max_flow_get_paths(self, start_v: Any, end_v: Any, edge_property: str = "weight") -> DeclarativeBase:
        return self.call("max_flow.get_paths", (start_v, end_v, edge_property))

    def node2vec_get_embeddings(
        self,
        is_directed: bool = False,
        p: Optional[Any] = 2,
        q: Optional[Any] = 0.5,
        num_walks: Optional[Any] = 4,
        walk_length: Optional[Any] = 5,
        vector_size: Optional[Any] = 100,
        alpha: Optional[Any] = 0.025,
        window: Optional[Any] = 5,
        min_count: Optional[Any] = 1,
        seed: Optional[Any] = 1,
        workers: Optional[Any] = 1,
        min_alpha: Optional[Any] = 0.0001,
        sg: Optional[Any] = 1,
        hs: Optional[Any] = 0,
        negative: Optional[Any] = 5,
        epochs: Optional[Any] = 5,
        edge_weight_property: Optional[Any] = "weight",
    ) -> DeclarativeBase:
        return self.call(
            "node2vec.get_embeddings",
            (
                is_directed,
                p,
                q,
                num_walks,
                walk_length,
                vector_size,
                alpha,
                window,
                min_count,
                seed,
                workers,
                min_alpha,
                sg,
                hs,
                negative,
                epochs,
                edge_weight_property,
            ),
        )

    def node2vec_help(self) -> DeclarativeBase:
        return self.call("node2vec.help")

    def node2vec_set_embeddings(
        self,
        is_directed: bool = False,
        p: Optional[Any] = 2,
        q: Optional[Any] = 0.5,
        num_walks: Optional[Any] = 4,
        walk_length: Optional[Any] = 5,
        vector_size: Optional[Any] = 100,
        alpha: Optional[Any] = 0.025,
        window: Optional[Any] = 5,
        min_count: Optional[Any] = 1,
        seed: Optional[Any] = 1,
        workers: Optional[Any] = 1,
        min_alpha: Optional[Any] = 0.0001,
        sg: Optional[Any] = 1,
        hs: Optional[Any] = 0,
        negative: Optional[Any] = 5,
        epochs: Optional[Any] = 5,
        edge_weight_property: Optional[Any] = "weight",
    ) -> DeclarativeBase:
        return self.call(
            "node2vec.set_embeddings",
            (
                is_directed,
                p,
                q,
                num_walks,
                walk_length,
                vector_size,
                alpha,
                window,
                min_count,
                seed,
                workers,
                min_alpha,
                sg,
                hs,
                negative,
                epochs,
                edge_weight_property,
            ),
        )

    def node2vec_online_get(self) -> DeclarativeBase:
        return self.call("node2vec_online.get")

    def node2vec_online_help(self) -> DeclarativeBase:
        return self.call("node2vec_online.help")

    def node2vec_online_reset(self) -> DeclarativeBase:
        return self.call("node2vec_online.reset")

    def node2vec_online_set_streamwalk_updater(
        self,
        half_life: int = 7200,
        max_length: int = 3,
        beta: float = 0.9,
        cutoff: int = 604800,
        sampled_walks: int = 4,
        full_walks: bool = False,
    ) -> DeclarativeBase:
        return self.call(
            "node2vec_online.set_streamwalk_updater", (half_life, max_length, beta, cutoff, sampled_walks, full_walks)
        )

    def node2vec_online_set_word2vec_learner(
        self,
        embedding_dimension: int = 128,
        learning_rate: float = 0.01,
        skip_gram: bool = True,
        negative_rate: float = 10,
        threads: Optional[int] = None,
    ) -> DeclarativeBase:
        return self.call(
            "node2vec_online.set_word2vec_learner",
            (embedding_dimension, learning_rate, skip_gram, negative_rate, threads),
        )

    def node2vec_online_update(self, edges: List[Any]) -> DeclarativeBase:
        return self.call("node2vec_online.update", (edges))

    def node_similarity_cosine(self, node1: Any, node2: Any, mode: str = "cartesian") -> DeclarativeBase:
        return self.call("node_similarity.cosine", (node1, node2, mode))

    def node_similarity_jaccard(self, node1: Any, node2: Any, mode: str = "cartesian") -> DeclarativeBase:
        return self.call("node_similarity.jaccard", (node1, node2, mode))

    def node_similarity_overlap(self, node1: Any, node2: Any, mode: str = "cartesian") -> DeclarativeBase:
        return self.call("node_similarity.overlap", (node1, node2, mode))

    def pagerank_get(
        self, max_iterations: int = 100, damping_factor: float = 0.85, stop_epsilon: float = 1e-05
    ) -> DeclarativeBase:
        return self.call("pagerank.get", (max_iterations, damping_factor, stop_epsilon))

    def pagerank_online_get(self) -> DeclarativeBase:
        return self.call("pagerank_online.get")

    def pagerank_online_reset(self) -> DeclarativeBase:
        return self.call("pagerank_online.reset")

    def pagerank_online_set(self, walks_per_node: int = 10, walk_stop_epsilon: float = 0.1) -> DeclarativeBase:
        return self.call("pagerank_online.set", (walks_per_node, walk_stop_epsilon))

    def pagerank_online_update(
        self,
        created_vertices: List[Any] = [],
        created_edges: List[Any] = [],
        deleted_vertices: List[Any] = [],
        deleted_edges: List[Any] = [],
    ) -> DeclarativeBase:
        return self.call("pagerank_online.update", (created_vertices, created_edges, deleted_vertices, deleted_edges))

    def rust_example_basic(self, input_string: str, optional_input_int: int = 0) -> DeclarativeBase:
        return self.call("rust_example.basic", (input_string, optional_input_int))

    def rust_example_test_procedure(self) -> DeclarativeBase:
        return self.call("rust_example.test_procedure")

    def set_cover_cp_solve(self, element_vertexes: List[Any], set_vertexes: List[Any]) -> DeclarativeBase:
        return self.call("set_cover.cp_solve", (element_vertexes, set_vertexes))

    def set_cover_greedy(self, element_vertexes: List[Any], set_vertexes: List[Any]) -> DeclarativeBase:
        return self.call("set_cover.greedy", (element_vertexes, set_vertexes))

    def tgn_get(self) -> DeclarativeBase:
        return self.call("tgn.get")

    def tgn_get_results(self) -> DeclarativeBase:
        return self.call("tgn.get_results")

    def tgn_predict_link_score(self, src: Any, dest: Any) -> DeclarativeBase:
        return self.call("tgn.predict_link_score", (src, dest))

    def tgn_reset(self) -> DeclarativeBase:
        return self.call("tgn.reset")

    def tgn_revert_from_database(self) -> DeclarativeBase:
        return self.call("tgn.revert_from_database")

    def tgn_save_tgn_params(self) -> DeclarativeBase:
        return self.call("tgn.save_tgn_params")

    def tgn_set_eval(self) -> DeclarativeBase:
        return self.call("tgn.set_eval")

    def tgn_set_params(self, params: Dict[str, Union[int, str]]) -> DeclarativeBase:
        return self.call("tgn.set_params", (params))

    def tgn_train_and_eval(self, num_epochs: int) -> DeclarativeBase:
        return self.call("tgn.train_and_eval", (num_epochs))

    def tgn_update(self, edges: List[Any]) -> DeclarativeBase:
        return self.call("tgn.update", (edges))

    def tsp_solve(self, points: List[Any], method: str = "1.5_approx") -> DeclarativeBase:
        return self.call("tsp.solve", (points, method))

    def union_find_connected(
        self, nodes1: Optional[Any], nodes2: Optional[Any], mode: str = "pairwise", update: bool = True
    ) -> DeclarativeBase:
        return self.call("union_find.connected", (nodes1, nodes2, mode, update))

    def uuid_generator_get(self) -> DeclarativeBase:
        return self.call("uuid_generator.get")

    def vrp_route(self, depot_node: Any, number_of_vehicles: Optional[int] = None) -> DeclarativeBase:
        return self.call("vrp.route", (depot_node, number_of_vehicles))

    def weakly_connected_components_get(self) -> DeclarativeBase:
        return self.call("weakly_connected_components.get")



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import socket
import subprocess
import time
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Union

import docker
import psutil

from gqlalchemy.exceptions import (
    GQLAlchemyWaitForConnectionError,
    GQLAlchemyWaitForDockerError,
    GQLAlchemyWaitForPortError,
)
from gqlalchemy.vendors.memgraph import Memgraph


MEMGRAPH_DEFAULT_BINARY_PATH = "/usr/lib/memgraph/memgraph"
MEMGRAPH_DEFAULT_PORT = 7687
MEMGRAPH_CONFIG_BOLT_PORT = "--bolt-port"
MEMGRAPH_CONFIG_BOLT_ADDRESS = "--bolt-address"
DOCKER_IMAGE_TAG_LATEST = "latest"
LOOPBACK_ADDRESS = "127.0.0.1"
WILDCARD_ADDRESS = "0.0.0.0"


class DockerImage(Enum):
    MEMGRAPH = "memgraph/memgraph"
    MAGE = "memgraph/memgraph-mage"


class DockerContainerStatus(Enum):
    EXITED = "exited"
    PAUSED = "paused"
    RESTARTING = "restarting"
    RUNNING = "running"


def wait_for_port(
    host: str = LOOPBACK_ADDRESS,
    port: int = MEMGRAPH_DEFAULT_PORT,
    delay: float = 0.01,
    timeout: float = 5.0,
    backoff: int = 2,
) -> None:
    """Wait for a TCP port to become available.

    Args:
        host: A string representing the IP address that is being checked.
        port: A string representing the port that is being checked.
        delay: A float that defines how long to wait between retries.
        timeout: A float that defines how long to wait for the port.
        backoff: An integer used for multiplying the delay.

    Raises:
      TimeoutError: Raises an error when the host and port are not accepting
        connections after the timeout period has passed.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as ex:
            time.sleep(delay)
            if time.perf_counter() - start_time >= timeout:
                raise GQLAlchemyWaitForPortError(port=port, host=host) from ex

        delay *= backoff


def wait_for_docker_container(
    container: "docker.Container", delay: float = 0.01, timeout: float = 5.0, backoff: int = 2
) -> None:
    """Wait for a Docker container to enter the status `running`.

    Args:
        container: The Docker container to wait for.
        delay: A float that defines how long to wait between retries.
        timeout: A float that defines how long to wait for the status.
        backoff: An integer used for multiplying the delay.

    Raises:
      TimeoutError: Raises an error when the container isn't running after the
        timeout period has passed.
    """
    start_time = time.perf_counter()
    while container.status != DockerContainerStatus.RUNNING.value:
        container.reload()
        time.sleep(delay)

        if time.perf_counter() - start_time >= timeout:
            raise GQLAlchemyWaitForDockerError

        delay *= backoff


class MemgraphInstance(ABC):
    def __init__(
        self,
        host: str = WILDCARD_ADDRESS,
        port: int = MEMGRAPH_DEFAULT_PORT,
        config: Dict[str, Union[str, int, bool]] = dict(),
    ) -> None:
        self.host = host
        self.port = port
        self.config = config
        self.proc_mg = None
        self.config[MEMGRAPH_CONFIG_BOLT_PORT] = self.port
        self.config[MEMGRAPH_CONFIG_BOLT_ADDRESS] = self.host
        self._memgraph = None

    @property
    def memgraph(self) -> Memgraph:
        if self._memgraph is None:
            self._memgraph = Memgraph(self.host, self.port)

        return self._memgraph

    def set_config(self, config: Dict[str, Union[str, int, bool]]) -> None:
        self.config.update(config)

    def connect(self) -> "Memgraph":
        if not self.is_running():
            raise GQLAlchemyWaitForConnectionError

        return self.memgraph

    def start_and_connect(self, restart: bool = False) -> "Memgraph":
        """Start the Memgraph instance and return the
          connection object.

        Attributes:
            restart: A bool indicating if the instance should be
              restarted if it's already running.
        """
        self.start(restart=restart)

        return self.connect()

    def start(self, restart: bool = False) -> None:
        """Start the Memgraph instance.

        Attributes:
            restart: A bool indicating if the instance should be
              restarted if it's already running.
        """
        if not restart and self.is_running():
            return

        self.stop()
        self._start_instance()

    def stop(self) -> Any:
        """Stop the Memgraph instance."""
        if not self.is_running():
            return

        self._stop_instance()

    @abstractmethod
    def is_running(self) -> bool:
        pass

    @abstractmethod
    def _start_instance(self) -> None:
        pass

    @abstractmethod
    def _stop_instance(self) -> Any:
        pass


class MemgraphInstanceBinary(MemgraphInstance):
    """A class for managing Memgraph instances started from binary files on Unix
    systems.

    Attributes:
      binary_path: A string representing the path to a Memgraph binary
        file.
      user: A string representing the user that should start the Memgraph
        process.
    """

    def __init__(self, binary_path: str = MEMGRAPH_DEFAULT_BINARY_PATH, user: str = "", **data) -> None:
        super().__init__(**data)
        self.binary_path = binary_path
        self.user = user

    def _start_instance(self) -> None:
        args_mg = f"{self.binary_path } " + (" ").join([f"{k}={v}" for k, v in self.config.items()])
        if self.user != "":
            args_mg = f"runuser -l {self.user} -c '{args_mg}'"

        self.proc_mg = subprocess.Popen(args_mg, shell=True)
        wait_for_port(self.host, self.port)

    def _stop_instance(self) -> None:
        procs = set()
        process = psutil.Process(self.proc_mg.pid)
        procs.add(process)
        for proc in process.children(recursive=True):
            procs.add(proc)
            proc.kill()

        process.kill()
        psutil.wait_procs(procs)

    def is_running(self) -> bool:
        """Check if the Memgraph instance is still running."""
        return self.proc_mg is not None and self.proc_mg.poll() is None


class MemgraphInstanceDocker(MemgraphInstance):
    """A class for managing Memgraph instances started in Docker containers.

    Attributes:
      docker_image: An enum representing the Docker image. Values:
        `DockerImage.MEMGRAPH` and `DockerImage.MAGE`.
      docker_image_tag: A string representing the tag of the Docker image.
    """

    def __init__(
        self, docker_image: DockerImage = DockerImage.MEMGRAPH, docker_image_tag: str = DOCKER_IMAGE_TAG_LATEST, **data
    ) -> None:
        super().__init__(**data)
        self.docker_image = docker_image
        self.docker_image_tag = docker_image_tag
        self._client = docker.from_env()
        self._container = None

    def _start_instance(self) -> None:
        self._container = self._client.containers.run(
            image=f"{self.docker_image.value}:{self.docker_image_tag}",
            command=f"{MEMGRAPH_DEFAULT_BINARY_PATH} {(' ').join([f'{k}={v}' for k, v in self.config.items()])}",
            detach=True,
            ports={f"{self.port}/tcp": self.port},
        )
        wait_for_docker_container(self._container)

    def _stop_instance(self) -> Dict:
        self._container.stop()

        return self._container.wait()

    def is_running(self) -> bool:
        """Check if the Memgraph instance is still running."""
        if self._container is None:
            return False

        self._container.reload()

        return self._container.status == DockerContainerStatus.RUNNING.value



# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
from abc import ABC, abstractmethod
from enum import Enum
from dataclasses import dataclass, field
from string import Template
from typing import List, Dict, Any, Optional, Union

import adlfs

try:
    import pyarrow.dataset as ds
except ModuleNotFoundError:
    ds = None
try:
    from pyarrow import fs
except ModuleNotFoundError:
    fs = None
from dacite import from_dict

from gqlalchemy import Memgraph
from gqlalchemy.models import (
    MemgraphIndex,
    MemgraphTrigger,
    TriggerEventObject,
    TriggerEventType,
    TriggerExecutionPhase,
)
from gqlalchemy.query_builders.memgraph_query_builder import Operator, QueryBuilder, Unwind
from gqlalchemy.transformations.importing.importer import Importer

NAME_MAPPINGS_KEY = "name_mappings"
ONE_TO_MANY_RELATIONS_KEY = "one_to_many_relations"
INDICES_KEY = "indices"
MANY_TO_MANY_RELATIONS_KEY = "many_to_many_relations"

FROM_NODE_VARIABLE_NAME = "from_node"
TO_NODE_VARIABLE_NAME = "to_node"

NODE_A = "a"
NODE_B = "b"

PARQUET_EXTENSION = "parquet"
CSV_EXTENSION = "csv"
ORC_EXTENSION = "orc"
IPC_EXTENSION = "ipc"
FEATHER_EXTENSION = "feather"
ARROW_EXTENSION = "arrow"

BLOB_ACCOUNT_NAME = "account_name"
BLOB_ACCOUNT_KEY = "account_key"
BLOB_SAS_TOKEN = "sas_token"
BLOB_CONTAINER_NAME_KEY = "container_name"

S3_REGION = "region"
S3_ACCESS_KEY = "access_key"
S3_SECRET_KEY = "secret_key"
S3_SESSION_TOKEN = "session_token"
S3_BUCKET_NAME_KEY = "bucket_name"

LOCAL_STORAGE_PATH = "local_storage_path"


@dataclass(frozen=True)
class ForeignKeyMapping:
    """Class that contains the full description of a single foreign key in a table.

    Attributes:
        column_name: Column name that holds the foreign key.
        reference_table: Name of a table from which the foreign key is taken.
        reference_key: Column name in the referenced table from which the foreign key is taken.
    """

    column_name: str
    reference_table: str
    reference_key: str


@dataclass(frozen=True)
class OneToManyMapping:
    """Class that holds the full description of a single one to many mapping in a table.

    Attributes:
        foreign_key: Foreign key used for mapping.
        label: Label which will be applied to the relationship created from this object.
        from_entity: Direction of the relationship created from the mapping object.
        parameters: Parameters that will be added to the relationship created from this object (Optional).
    """

    foreign_key: ForeignKeyMapping
    label: str
    from_entity: bool = False
    parameters: Optional[Dict[str, str]] = None


@dataclass(frozen=True)
class ManyToManyMapping:
    """Class that holds the full description of a single many to many mapping in a table.
    Many to many mapping is intended to be used in case of associative tables.

    Attributes:
        foreign_key_from: Describes the source of the relationship.
        foreign_key_to: Describes the destination of the relationship.
        label: Label to be applied to the newly created relationship.
        parameters: Parameters that will be added to the relationship created from this object (Optional).
    """

    foreign_key_from: ForeignKeyMapping
    foreign_key_to: ForeignKeyMapping
    label: str
    parameters: Optional[Dict[str, str]] = None


Mapping = Union[List[OneToManyMapping], ManyToManyMapping]


@dataclass
class TableMapping:
    """Class that holds the full description of all of the mappings for a single table.

    Attributes:
        table_name: Name of the table.
        mapping: All of the mappings in the table (Optional).
        indices: List of the indices to be created for this table (Optional).
    """

    table_name: str
    mapping: Optional[Mapping] = None
    indices: Optional[List[str]] = None


@dataclass(frozen=True)
class NameMappings:
    """Class that contains new label name and all of the column name mappings for a single table.

    Attributes:
        label: New label (Optional).
        column_names_mapping: Dictionary containing key-value pairs in form ("column name", "property name") (Optional).
    """

    label: Optional[str] = None
    column_names_mapping: Dict[str, str] = field(default_factory=dict)

    def get_property_name(self, column_name: str):
        return self.column_names_mapping.get(column_name, column_name)


class NameMapper:
    """Class that holds all name mappings for all of the collections."""

    def __init__(self, mappings: Dict[str, Any]) -> None:
        self._name_mappings: Dict[str, NameMappings] = {k: NameMappings(**v) for k, v in mappings.items()}

    def get_label(self, collection_name: str) -> str:
        """Returns label for given collection.

        Args:
            collection_name: Original collection name.
        """
        label = self._name_mappings[collection_name].label

        return label if label is not None else collection_name

    def get_property_name(self, collection_name: str, column_name: str) -> str:
        """Returns property name for column from collection.

        Args:
            collection_name: Original collection name.
            column_name: Original column name.
        """
        return self._name_mappings[collection_name].get_property_name(column_name=column_name)


class FileSystemHandler(ABC):
    """Abstract class for defining FileSystemHandler.

    Inherit this class, define a custom data source and initialize the
    connection.
    """

    def __init__(self, fs: Any) -> None:
        super().__init__()
        self._fs = fs

    @property
    def fs(self):
        return self._fs

    @abstractmethod
    def get_path(self, collection_name: str) -> str:
        """Returns complete path in specific file system. Used to read the file system
        for a specific file.
        """
        pass


class S3FileSystemHandler(FileSystemHandler):
    """Handles connection to Amazon S3 service via PyArrow."""

    def __init__(self, bucket_name: str, **kwargs):
        """Initializes connection and data bucket.

        Args:
            bucket_name: Name of the bucket on S3 from which to read the data

        Kwargs:
            access_key: S3 access key.
            secret_key: S3 secret key.
            region: S3 region.
            session_token: S3 session token (Optional).

        Raises:
            KeyError: kwargs doesn't contain necessary fields.
        """
        if S3_ACCESS_KEY not in kwargs:
            raise KeyError(f"{S3_ACCESS_KEY} is needed to connect to S3 storage")
        if S3_SECRET_KEY not in kwargs:
            raise KeyError(f"{S3_SECRET_KEY} is needed to connect to S3 storage")

        if fs is None:
            raise ModuleNotFoundError("No module named 'pyarrow'")

        super().__init__(fs=fs.S3FileSystem(**kwargs))
        self._bucket_name = bucket_name

    def get_path(self, collection_name: str) -> str:
        """Get file path in file system.

        Args:
            collection_name: Name of the file to read.
        """
        return f"{self._bucket_name}/{collection_name}"


class AzureBlobFileSystemHandler(FileSystemHandler):
    """Handles connection to Azure Blob service via adlfs package."""

    def __init__(self, container_name: str, **kwargs) -> None:
        """Initializes connection and data container.

        Args:
            container_name: Name of the Blob container storing data.

        Kwargs:
            account_name: Account name from Azure Blob.
            account_key: Account key for Azure Blob (Optional - if using sas_token).
            sas_token: Shared access signature token for authentification (Optional).

        Raises:
            KeyError: kwargs doesn't contain necessary fields.
        """
        if BLOB_ACCOUNT_KEY not in kwargs and BLOB_SAS_TOKEN not in kwargs:
            raise KeyError(f"{BLOB_ACCOUNT_KEY} or {BLOB_SAS_TOKEN} is needed to connect to Blob storage")
        if BLOB_ACCOUNT_NAME not in kwargs:
            raise KeyError(f"{BLOB_ACCOUNT_NAME} is needed to connect to Blob storage")

        super().__init__(fs=adlfs.AzureBlobFileSystem(**kwargs))
        self._container_name = container_name

    def get_path(self, collection_name: str) -> str:
        """Get file path in file system.

        Args:
            collection_name: Name of the file to read.
        """
        return f"{self._container_name}/{collection_name}"


class LocalFileSystemHandler(FileSystemHandler):
    """Handles a local filesystem."""

    def __init__(self, path: str) -> None:
        """Initializes an fsspec local file system and sets path to data.

        Args:
            path: path to the local storage location.
        """
        if fs is None:
            raise ModuleNotFoundError("No module named 'pyarrow'")

        super().__init__(fs=fs.LocalFileSystem())
        self._path = path

    def get_path(self, collection_name: str) -> str:
        """Get file path in the local file system.

        Args:
            collection_name: Name of the file to read.
        """
        return f"{self._path}/{collection_name}"


class DataLoader(ABC):
    """Implements loading of a data type from file system service to TableToGraphImporter."""

    def __init__(self, file_extension: str, file_system_handler: FileSystemHandler) -> None:
        """
        Args:
            file_extension: File format to be read.
            file_system_handler: Object for handling of the file system service.
        """
        super().__init__()
        self._file_extension = file_extension
        self._file_system_handler = file_system_handler

    @abstractmethod
    def load_data(self, collection_name: str, is_cross_table: bool = False) -> None:
        """Override this method in the derived class. Intended to be used for reading data from data format.

        Args:
            collection_name: Name of the file to read.
            is_cross_table: Indicate whether or not the collection contains associative table (default=False).

        Raises:
            NotImplementedError: The method is not implemented in the extended class.
        """
        raise NotImplementedError("Subclasses must override load_data() for use in TableToGraphImporter")


class PyArrowFileTypeEnum(Enum):
    """Enumerates file types supported by PyArrow"""

    Default = 1
    Parquet = 2
    CSV = 3
    ORC = 4
    Feather = 5


class PyArrowDataLoader(DataLoader):
    """Loads data using PyArrow.

    PyArrow currently supports "parquet", "ipc"/"arrow"/"feather", "csv",
    and "orc", see pyarrow.dataset.dataset for up-to-date info.
    ds.dataset in load_data accepts any fsspec subclass, making this DataLoader
    compatible with fsspec-compatible filesystems.
    """

    def __init__(
        self,
        file_extension_enum: PyArrowFileTypeEnum,
        file_system_handler: FileSystemHandler,
    ) -> None:
        """
        Args:
            file_extension_enum: The file format to be read.
            file_system_handler: Object for handling of the file system service.
        """
        super().__init__(file_extension=file_extension_enum.name.lower(), file_system_handler=file_system_handler)

    def load_data(
        self, collection_name: str, is_cross_table: bool = False, columns: Optional[List[str]] = None
    ) -> None:
        """Generator for loading data.

        Args:
            collection_name: Name of the file to read.
            is_cross_table: Flag signifying whether it is a cross table.
            columns: Table columns to read.
        """
        source = self._file_system_handler.get_path(f"{collection_name}.{self._file_extension}")
        print("Loading data from " + ("cross " if is_cross_table else "") + f"table {source}...")

        if ds is None:
            raise ModuleNotFoundError("No module named 'pyarrow'")

        dataset = ds.dataset(source=source, format=self._file_extension, filesystem=self._file_system_handler.fs)

        for batch in dataset.to_batches(
            columns=columns,
        ):
            for batch_item in batch.to_pylist():
                yield batch_item

        print("Data loaded.")


class TableToGraphImporter(Importer):
    """Implements translation of table data to graph data, and imports it to Memgraph."""

    _DIRECTION = {
        True: (NODE_A, NODE_B),
        False: (NODE_B, NODE_A),
    }

    _TriggerQueryTemplate = Template(
        Unwind(list_expression="createdVertices", variable="$node_a")
        .with_(results={"$node_a": ""})
        .where(item="$node_a", operator=Operator.LABEL_FILTER, expression="$label_2")
        .match()
        .node(labels="$label_1", variable="$node_b")
        .where(item="$node_b.$property_1", operator=Operator.EQUAL, expression="$node_a.$property_2")
        .create()
        .node(variable="$from_node")
        .to(relationship_type="$relationship_type")
        .node(variable="$to_node")
        .construct_query()
    )

    @staticmethod
    def _create_trigger_cypher_query(
        label1: str, label2: str, property1: str, property2: str, relationship_type: str, from_entity: bool
    ) -> str:
        """Creates a Cypher query for the translation trigger.

        Args:
            label1: Label of the first node.
            label2: Label of the second node.
            property1: Property of the first node.
            property2: Property of the second node.
            relationship_type: Label for the relationship that the trigger creates.
            from_entity: Indicate whether the relationship goes from or to the first entity.
        """
        from_node, to_node = TableToGraphImporter._DIRECTION[from_entity]

        return TableToGraphImporter._TriggerQueryTemplate.substitute(
            node_a=NODE_A,
            node_b=NODE_B,
            label_1=label1,
            label_2=label2,
            property_1=property1,
            property_2=property2,
            from_node=from_node,
            to_node=to_node,
            relationship_type=relationship_type,
        )

    def __init__(
        self,
        data_loader: DataLoader,
        data_configuration: Dict[str, Any],
        memgraph: Optional[Memgraph] = None,
    ) -> None:
        """
        Args:
            data_loader: Object for loading data.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
        """
        self._data_loader: DataLoader = data_loader
        self._memgraph: Memgraph = memgraph if memgraph is not None else Memgraph()

        self.__load_configuration(data_configuration=data_configuration)

    def translate(self, drop_database_on_start: bool = True) -> None:
        """Performs the translations.

        Args:
            drop_database_on_start: Indicate whether or not the database should be dropped prior to the start of the translations.
        """
        if drop_database_on_start:
            self._memgraph.drop_database()
            self._memgraph.drop_indexes()
            self._memgraph.drop_triggers()

        self._create_indexes()
        self._create_triggers()

        self._load_nodes()
        self._load_cross_relationships()

    def _load_nodes(self) -> None:
        """Reads all of the data from the single table in the data source, translates it, and writes it to Memgraph."""
        for one_to_many_mapping in self._one_to_many_mappings:
            collection_name = one_to_many_mapping.table_name
            for row in self._data_loader.load_data(collection_name=collection_name):
                self._save_row_as_node(label=collection_name, row=row)

    def _load_cross_relationships(self) -> None:
        """Reads all of the data from the single associative table in the data source, translates it, and writes it to Memgraph."""
        for many_to_many_mapping in self._many_to_many_mappings:
            mapping_from = many_to_many_mapping.mapping.foreign_key_from
            mapping_to = many_to_many_mapping.mapping.foreign_key_to

            for row in self._data_loader.load_data(
                collection_name=many_to_many_mapping.table_name, is_cross_table=True
            ):
                self._save_row_as_relationship(
                    collection_name_from=mapping_from.reference_table,
                    collection_name_to=mapping_to.reference_table,
                    property_from=mapping_from.reference_key,
                    property_to=mapping_to.reference_key,
                    relation_label=many_to_many_mapping.mapping.label,
                    row=row,
                )

    def _create_triggers(self) -> None:
        """Creates all of the Triggers in Memgraph.

        Triggers are used as a part of speeding up the translation. Since nodes
        and relationships are written in one go, foreign keys that are represented
        as relationships might not yet be present in memgraph. When they do appear,
        triggers make sure to create the relationship at that point in time, rather
        than having hanging relationship.
        """
        for one_to_many_mapping in self._one_to_many_mappings:
            label1 = self._name_mapper.get_label(collection_name=one_to_many_mapping.table_name)
            for mapping in one_to_many_mapping.mapping:
                property1 = self._name_mapper.get_property_name(
                    collection_name=one_to_many_mapping.table_name, column_name=mapping.foreign_key.column_name
                )
                label2 = self._name_mapper.get_label(collection_name=mapping.foreign_key.reference_table)
                property2 = self._name_mapper.get_property_name(
                    collection_name=one_to_many_mapping.table_name, column_name=mapping.foreign_key.reference_key
                )
                relationship_type = mapping.label
                from_entity = mapping.from_entity

                self._create_trigger(
                    label1=label1,
                    label2=label2,
                    property1=property1,
                    property2=property2,
                    relationship_type=relationship_type,
                    from_entity=from_entity,
                )
                self._create_trigger(
                    label1=label2,
                    label2=label1,
                    property1=property2,
                    property2=property1,
                    relationship_type=relationship_type,
                    from_entity=not from_entity,
                )

    def _create_trigger(
        self, label1: str, label2: str, property1: str, property2: str, relationship_type: str, from_entity: bool
    ) -> None:
        """Creates a translation trigger in Memgraph.

        Args:
            label1: Label of the first node.
            label2: Label of the second node.
            property1: Property of the first node.
            property2: Property of the second node.
            relationship_type: Label for the relationship that the trigger creates.
            from_entity: Indicate whether the relationship goes from or to the first entity.
        """
        trigger_name = "__".join([label1, property1, label2, property2])

        trigger = MemgraphTrigger(
            name=trigger_name,
            event_type=TriggerEventType.CREATE,
            event_object=TriggerEventObject.NODE,
            execution_phase=TriggerExecutionPhase.BEFORE,
            statement=TableToGraphImporter._create_trigger_cypher_query(
                label1, label2, property1, property2, relationship_type, from_entity
            ),
        )

        self._memgraph.create_trigger(trigger)

        print(f"Created trigger {trigger_name}")

    def _create_indexes(self) -> None:
        """Creates indices in Memgraph."""
        for one_to_many_mapping in self._one_to_many_mappings:
            collection_name = self._name_mapper.get_label(collection_name=one_to_many_mapping.table_name)
            for index in one_to_many_mapping.indices:
                new_index = self._name_mapper.get_property_name(
                    collection_name=one_to_many_mapping.table_name, column_name=index
                )
                self._memgraph.create_index(index=MemgraphIndex(collection_name, new_index))
                print(f"Created index for {collection_name} on {new_index}")

    def _save_row_as_node(
        self,
        label: str,
        row: Dict[str, Any],
    ) -> None:
        """Translates a row to a node and writes it to Memgraph.

        Args:
            label: Original label of the new node.
            row: The row that should be saved to Memgraph as a node.
        """
        (
            QueryBuilder(connection=self._memgraph)
            .create()
            .node(
                labels=self._name_mapper.get_label(collection_name=label),
                **{
                    self._name_mapper.get_property_name(collection_name=label, column_name=k): v for k, v in row.items()
                },
            )
            .execute()
        )

    def _save_row_as_relationship(
        self,
        collection_name_from: str,
        collection_name_to: str,
        property_from: str,
        property_to: str,
        relation_label: str,
        row: Dict[str, Any],
    ) -> None:
        """Translates a row to a relationship and writes it to Memgraph.

        Args:
            collection_name_from: Collection name of the source node.
            collection_name_to: Collection name of the destination node.
            property_from: Property of the source node.
            property_to: Property of the destination node.
            relation_label: Label for the relationship.
            row: The row to be translated.
        """
        (
            QueryBuilder(connection=self._memgraph)
            .match()
            .node(
                labels=self._name_mapper.get_label(collection_name=collection_name_from),
                variable=NODE_A,
                **{
                    self._name_mapper.get_property_name(
                        collection_name=collection_name_from, column_name=property_from
                    ): row[property_from]
                },
            )
            .match()
            .node(
                labels=self._name_mapper.get_label(collection_name=collection_name_to),
                variable=NODE_B,
                **{
                    self._name_mapper.get_property_name(
                        collection_name=collection_name_to, column_name=property_to
                    ): row[property_to]
                },
            )
            .create()
            .node(variable=NODE_A)
            .to(relation_label)
            .node(variable=NODE_B)
            .execute()
        )

    def __load_configuration(self, data_configuration: Dict[str, Any]) -> None:
        """Loads all of the configuration.

        Args:
            data_configuration: instructions to translate table to graph.
        """
        self.__load_name_mappings(data_configuration.get(NAME_MAPPINGS_KEY, {}))
        self.__load_one_to_many_mappings_and_indices(
            data_configuration[ONE_TO_MANY_RELATIONS_KEY], data_configuration.get(INDICES_KEY, {})
        )
        self.__load_many_to_many_mappings(data_configuration.get(MANY_TO_MANY_RELATIONS_KEY, {}))

    def __load_name_mappings(self, name_mappings: Dict[str, Any]) -> None:
        """Loads name mappings from the configuration."""
        self._name_mapper = NameMapper(mappings=name_mappings)

    def __load_one_to_many_mappings_and_indices(
        self, one_to_many_configuration: Dict[str, List[str]], indices: Dict[str, List[str]]
    ) -> None:
        """Loads One To Many Mappings and indices from the configuration."""
        self._one_to_many_mappings = [
            TableMapping(
                table_name=table_name,
                mapping=[from_dict(data_class=OneToManyMapping, data=relation) for relation in relations],
                indices=indices.get(table_name, {}),
            )
            for table_name, relations in one_to_many_configuration.items()
        ]

    def __load_many_to_many_mappings(self, many_to_many_configuration: Dict[str, Any]) -> None:
        """Loads Many To Many Mappings from the configuration."""
        self._many_to_many_mappings = [
            TableMapping(table_name=table_name, mapping=from_dict(data_class=ManyToManyMapping, data=relations))
            for table_name, relations in many_to_many_configuration.items()
        ]


class PyArrowImporter(TableToGraphImporter):
    """TableToGraphImporter wrapper for use with PyArrow for reading data."""

    def __init__(
        self,
        file_system_handler: str,
        file_extension_enum: PyArrowFileTypeEnum,
        data_configuration: Dict[str, Any],
        memgraph: Optional[Memgraph] = None,
    ) -> None:
        """
        Args:
            file_system_handler: File system to read from.
            file_extension_enum: File format to be read.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).

        Raises:
            ValueError: PyArrow doesn't support ORC on Windows.
        """
        if file_extension_enum == PyArrowFileTypeEnum.ORC and platform.system() == "Windows":
            raise ValueError("ORC filetype is currently not supported by PyArrow on Windows")

        super().__init__(
            data_loader=PyArrowDataLoader(
                file_extension_enum=file_extension_enum, file_system_handler=file_system_handler
            ),
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class PyArrowS3Importer(PyArrowImporter):
    """PyArrowImporter wrapper for use with the Amazon S3 File System."""

    def __init__(
        self,
        bucket_name: str,
        file_extension_enum: PyArrowFileTypeEnum,
        data_configuration: Dict[str, Any],
        memgraph: Optional[Memgraph] = None,
        **kwargs,
    ) -> None:
        """
        Args:
            bucket_name: Name of the bucket in S3 to read from.
            file_extension_enum: File format to be read.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for S3FileSystem.
        """
        super().__init__(
            file_system_handler=S3FileSystemHandler(bucket_name=bucket_name, **kwargs),
            file_extension_enum=file_extension_enum,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class PyArrowAzureBlobImporter(PyArrowImporter):
    """PyArrowImporter wrapper for use with the Azure Blob File System."""

    def __init__(
        self,
        container_name: str,
        file_extension_enum: PyArrowFileTypeEnum,
        data_configuration: Dict[str, Any],
        memgraph: Optional[Memgraph] = None,
        **kwargs,
    ) -> None:
        """
        Args:
            container_name: Name of the container in Azure Blob to read from.
            file_extension_enum: File format to be read.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for AzureBlobFileSystem.
        """
        super().__init__(
            file_system_handler=AzureBlobFileSystemHandler(container_name=container_name, **kwargs),
            file_extension_enum=file_extension_enum,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class PyArrowLocalFileSystemImporter(PyArrowImporter):
    """PyArrowImporter wrapper for use with the Local File System."""

    def __init__(
        self,
        path: str,
        file_extension_enum: PyArrowFileTypeEnum,
        data_configuration: Dict[str, Any],
        memgraph: Optional[Memgraph] = None,
    ) -> None:
        """
        Args:
            path: Full path to the directory to read from.
            file_extension_enum: File format to be read.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
        """
        super().__init__(
            file_system_handler=LocalFileSystemHandler(path=path),
            file_extension_enum=file_extension_enum,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class ParquetS3FileSystemImporter(PyArrowS3Importer):
    """PyArrowS3Importer wrapper for use with the S3 file system and the parquet file type."""

    def __init__(
        self, bucket_name: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            bucket_name: Name of the bucket in S3 to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for S3FileSystem.
        """
        super().__init__(
            bucket_name=bucket_name,
            file_extension_enum=PyArrowFileTypeEnum.Parquet,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class CSVS3FileSystemImporter(PyArrowS3Importer):
    """PyArrowS3Importer wrapper for use with the S3 file system and the CSV file type."""

    def __init__(
        self, bucket_name: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            bucket_name: Name of the bucket in S3 to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for S3FileSystem.
        """
        super().__init__(
            bucket_name=bucket_name,
            file_extension_enum=PyArrowFileTypeEnum.CSV,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class ORCS3FileSystemImporter(PyArrowS3Importer):
    """PyArrowS3Importer wrapper for use with the S3 file system and the ORC file type."""

    def __init__(
        self, bucket_name: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            bucket_name: Name of the bucket in S3 to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for S3FileSystem.
        """
        super().__init__(
            bucket_name=bucket_name,
            file_extension_enum=PyArrowFileTypeEnum.ORC,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class FeatherS3FileSystemImporter(PyArrowS3Importer):
    """PyArrowS3Importer wrapper for use with the S3 file system and the feather file type."""

    def __init__(
        self, bucket_name: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            bucket_name: Name of the bucket in S3 to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for S3FileSystem.
        """
        super().__init__(
            bucket_name=bucket_name,
            file_extension_enum=PyArrowFileTypeEnum.Feather,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class ParquetAzureBlobFileSystemImporter(PyArrowAzureBlobImporter):
    """PyArrowAzureBlobImporter wrapper for use with the Azure Blob file system and the parquet file type."""

    def __init__(
        self, container_name: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            container_name: Name of the container in Azure Blob storage to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for AzureBlobFileSystem.
        """
        super().__init__(
            container_name=container_name,
            file_extension_enum=PyArrowFileTypeEnum.Parquet,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class CSVAzureBlobFileSystemImporter(PyArrowAzureBlobImporter):
    """PyArrowAzureBlobImporter wrapper for use with the Azure Blob file system and the CSV file type."""

    def __init__(
        self, container_name: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            container_name: Name of the container in Azure Blob storage to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for AzureBlobFileSystem.
        """
        super().__init__(
            container_name=container_name,
            file_extension_enum=PyArrowFileTypeEnum.CSV,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class ORCAzureBlobFileSystemImporter(PyArrowAzureBlobImporter):
    """PyArrowAzureBlobImporter wrapper for use with the Azure Blob file system and the CSV file type."""

    def __init__(
        self, container_name, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            container_name: Name of the container in Blob storage to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for AzureBlobFileSystem.
        """
        super().__init__(
            container_name=container_name,
            file_extension_enum=PyArrowFileTypeEnum.ORC,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class FeatherAzureBlobFileSystemImporter(PyArrowAzureBlobImporter):
    """PyArrowAzureBlobImporter wrapper for use with the Azure Blob file system and the Feather file type."""

    def __init__(
        self, container_name, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None, **kwargs
    ) -> None:
        """
        Args:
            container_name: Name of the container in Blob storage to read from.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for AzureBlobFileSystem.
        """
        super().__init__(
            container_name=container_name,
            file_extension_enum=PyArrowFileTypeEnum.Feather,
            data_configuration=data_configuration,
            memgraph=memgraph,
            **kwargs,
        )


class ParquetLocalFileSystemImporter(PyArrowLocalFileSystemImporter):
    """PyArrowLocalFileSystemImporter wrapper for use with the local file system and the parquet file type."""

    def __init__(self, path: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None) -> None:
        """
        Args:
            path: Full path to directory.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for LocalFileSystem.
        """
        super().__init__(
            path=path,
            file_extension_enum=PyArrowFileTypeEnum.Parquet,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class CSVLocalFileSystemImporter(PyArrowLocalFileSystemImporter):
    """PyArrowLocalFileSystemImporter wrapper for use with the local file system and the CSV file type."""

    def __init__(self, path: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None) -> None:
        """
        Args:
            path: Full path to directory.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for LocalFileSystem.
        """
        super().__init__(
            path=path,
            file_extension_enum=PyArrowFileTypeEnum.CSV,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class ORCLocalFileSystemImporter(PyArrowLocalFileSystemImporter):
    """PyArrowLocalFileSystemImporter wrapper for use with the local file system and the ORC file type."""

    def __init__(self, path: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None) -> None:
        """
        Args:
            path: Full path to directory.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for LocalFileSystem.
        """
        super().__init__(
            path=path,
            file_extension_enum=PyArrowFileTypeEnum.ORC,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )


class FeatherLocalFileSystemImporter(PyArrowLocalFileSystemImporter):
    """PyArrowLocalFileSystemImporter wrapper for use with the local file system and the Feather/IPC/Arrow file type."""

    def __init__(self, path: str, data_configuration: Dict[str, Any], memgraph: Optional[Memgraph] = None) -> None:
        """
        Args:
            path: Full path to directory.
            data_configuration: Configuration for the translations.
            memgraph: Connection to Memgraph (Optional).
            **kwargs: Specified for LocalFileSystem.
        """
        super().__init__(
            path=path,
            file_extension_enum=PyArrowFileTypeEnum.Feather,
            data_configuration=data_configuration,
            memgraph=memgraph,
        )



# Copyright (c) 2016-2023 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gqlalchemy.exceptions import raise_if_not_imported
import gqlalchemy.memgraph_constants as mg_consts
from gqlalchemy.transformations.export.transporter import Transporter
from gqlalchemy.transformations.graph_type import GraphType

try:
    from gqlalchemy.transformations.translators.dgl_translator import DGLTranslator
except ModuleNotFoundError:
    DGLTranslator = None

from gqlalchemy.transformations.translators.nx_translator import NxTranslator

try:
    from gqlalchemy.transformations.translators.pyg_translator import PyGTranslator
except ModuleNotFoundError:
    PyGTranslator = None


class GraphTransporter(Transporter):
    """Here is a possible example for using this module:
    >>> transporter = GraphTransporter("dgl")
    graph = transporter.export()
    """

    def __init__(
        self,
        graph_type: str,
        host: str = mg_consts.MG_HOST,
        port: int = mg_consts.MG_PORT,
        username: str = mg_consts.MG_USERNAME,
        password: str = mg_consts.MG_PASSWORD,
        encrypted: bool = mg_consts.MG_ENCRYPTED,
        client_name: str = mg_consts.MG_CLIENT_NAME,
        lazy: bool = mg_consts.MG_LAZY,
    ) -> None:
        """Initializes GraphTransporter. It is used for converting Memgraph graph to the specific graph type offered by some Python package (PyG, DGL, NX...)
        Here is a possible example for using this module:
        >>> transporter = GraphTransporter("dgl")
        graph = transporter.export()
        Args:
            graph_type: dgl, pyg or nx
        """
        super().__init__()
        self.graph_type = graph_type.upper()
        if self.graph_type == GraphType.DGL.name:
            raise_if_not_imported(dependency=DGLTranslator, dependency_name="dgl")
            self.translator = DGLTranslator(host, port, username, password, encrypted, client_name, lazy)
        elif self.graph_type == GraphType.PYG.name:
            raise_if_not_imported(dependency=PyGTranslator, dependency_name="torch_geometric")
            self.translator = PyGTranslator(host, port, username, password, encrypted, client_name, lazy)
        elif self.graph_type == GraphType.NX.name:
            self.translator = NxTranslator(host, port, username, password, encrypted, client_name, lazy)
        else:
            raise ValueError("Unknown export option. Currently supported are DGL, PyG and Networkx.")

    def export(self):
        """Creates graph instance for the wanted export option."""
        return self.translator.get_instance()



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import json
from datetime import datetime

import mgp

API_URLS = {}
ORDER_NUMBER = 0


@mgp.read_proc
def create_push_stream(
    context: mgp.ProcCtx,
    stream_name: str,
    api_url: str,
) -> mgp.Record(created=bool):
    global API_URLS

    if not isinstance(stream_name, str):
        raise TypeError("Invalid type on first argument!")

    if not isinstance(api_url, str):
        raise TypeError("Invalid type on second argument!")

    API_URLS[stream_name] = api_url

    return mgp.Record(created=True)


@mgp.read_proc
def show_streams(
    context: mgp.ProcCtx,
) -> mgp.Record(name=str, api_url=str):
    records = []

    for k, v in API_URLS.items():
        records.append(mgp.Record(name=k, api_url=v))

    return records


@mgp.read_proc
def push(
    context: mgp.ProcCtx,
    stream_name: str,
    payload: mgp.Map,
) -> mgp.Record(status=str):

    if not isinstance(stream_name, str):
        raise TypeError("Invalid type on first argument!")

    if not isinstance(payload, dict):
        raise TypeError("Invalid type on second argument!")

    if stream_name not in API_URLS:
        raise Exception("Power BI stream not defined!")

    api_url = API_URLS[stream_name]

    message = ""
    if isinstance(payload, dict):
        message = payload
    elif isinstance(payload, mgp.Vertex) or isinstance(payload, mgp.Edge):
        message = {x.name: x.value for x in payload.properties.items()}
    else:
        raise Exception("Can't have message type other than Map / Vertex / Edge")

    for k, v in message.items():
        if isinstance(v, datetime):
            message[k] = datetime.strftime(v, "%Y-%m-%dT%H:%M:%S")

    headers = {"Content-Type": "application/json"}

    try:
        response = requests.request(method="POST", url=api_url, headers=headers, data=json.dumps(message))
    except Exception as e:
        raise Exception(f"Error happened while sending results! {e}")

    return mgp.Record(status=str(response.status))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gqlalchemy',
 'gqlalchemy.graph_algorithms',
 'gqlalchemy.query_builders',
 'gqlalchemy.query_modules',
 'gqlalchemy.query_modules.push_streams',
 'gqlalchemy.transformations',
 'gqlalchemy.transformations.export',
 'gqlalchemy.transformations.importing',
 'gqlalchemy.transformations.translators',
 'gqlalchemy.vendors']

package_data = \
{'': ['*']}

install_requires = \
['adlfs>=2022.2.0,<2023.0.0',
 'dacite>=1.6.0,<2.0.0',
 'docker>=5.0.3,<6.0.0',
 'neo4j>=4.4.3,<5.0.0',
 'networkx==3.1',
 'numpy>=1.24.1,<2.0.0',
 'platformdirs==3.8.0',
 'psutil>=5.9.0,<6.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'pymgclient==1.3.1']

extras_require = \
{'all': ['pyarrow>=9.0.0,<10.0.0',
         'torch>=1.13.1,<2.0.0',
         'dgl>=0.9.1,<0.10.0'],
 'arrow': ['pyarrow>=9.0.0,<10.0.0'],
 'dgl': ['torch>=1.13.1,<2.0.0', 'dgl>=0.9.1,<0.10.0'],
 'torch-pyg': ['torch>=1.13.1,<2.0.0']}

setup_kwargs = {
    'name': 'gqlalchemy-tailor',
    'version': '1.4.1',
    'description': 'GQLAlchemy is a library developed to assist with writing and running queries in Memgraph. Upgraded networkx to 3.1.0.',
    'long_description': '# GQLAlchemy\n\n\n<p>\n    <a href="https://github.com/memgraph/gqlalchemy/actions"><img src="https://github.com/memgraph/gqlalchemy/workflows/Build%20and%20Test/badge.svg" /></a>\n    <a href="https://github.com/memgraph/gqlalchemy/blob/main/LICENSE"><img src="https://img.shields.io/github/license/memgraph/gqlalchemy" /></a>\n    <a href="https://pypi.org/project/gqlalchemy"><img src="https://img.shields.io/pypi/v/gqlalchemy" /></a>\n    <a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n    <a href="https://memgraph.com/docs/gqlalchemy" alt="Documentation"><img src="https://img.shields.io/badge/documentation-GQLAlchemy-orange" /></a>\n    <a href="https://github.com/memgraph/gqlalchemy/stargazers" alt="Stargazers"><img src="https://img.shields.io/github/stars/memgraph/gqlalchemy?style=social" /></a>\n</p>\n\n**GQLAlchemy** is a fully open-source Python library and **Object Graph Mapper** (OGM) - a link between graph database objects and Python objects.\n\nAn Object Graph Mapper or OGM provides a developer-friendly workflow that allows for writing object-oriented notation to communicate with graph databases. Instead of writing Cypher queries, you will be able to write object-oriented code, which the OGM will automatically translate into Cypher queries.\n\n## Installation\n\nGQLAlchemy is built on top of Memgraph\'s low-level Python client `pymgclient`\n([PyPI](https://pypi.org/project/pymgclient/) /\n[Documentation](https://memgraph.github.io/pymgclient/) /\n[GitHub](https://github.com/memgraph/pymgclient)). \n\nTo install GQLAlchemy, you first need to install `pymgclient` [build prerequisites](https://memgraph.github.io/pymgclient/introduction.html#build-prerequisites).\n\nAfter you have installed the prerequisites, run the following command:\n\n`pip install gqlalchemy`\n\nWith the above command, you get the basic GQLAlchemy capabilities. To add additional import/export capabilities, install GQLAlchemy with one of the following commands:\n\n- `pip install gqlalchemy[arrow]` # Support for the CSV, Parquet, ORC and IPC/Feather/Arrow formats\n- `pip install gqlalchemy[dgl]` # DGL support (includes PyTorch)\n\n- `pip install gqlalchemy[all]` # All of the above\n\nIf you intend to use GQLAlchemy with PyTorch Geometric support, that library must be installed manually:\n\n```bash\npip install gqlalchemy[torch_pyg] # prerequisite\npip install torch-scatter torch-sparse torch-cluster torch-spline-conv torch-geometric -f https://data.pyg.org/whl/torch-1.13.0+cpu.html"\n```\n\nIf you are using [Conda](https://docs.conda.io/en/latest/) for Python environment management, you can install GQLAlchemy through pip.\n\n## Build & Test\n\nThe project uses [Poetry](https://python-poetry.org/) to build the library. Clone or download the [GQLAlchemy source code](https://github.com/memgraph/gqlalchemy) locally and run the following command to build it from source with Poetry:\n\n```bash\npoetry install --all-extras\n```\n\nThe ``poetry install --all-extras`` command installs GQLAlchemy with all extras\n(optional dependencies). Alternatively, you can use the ``-E`` option to define\nwhat extras to install:\n\n```bash\npoetry install # No extras\n\npoetry install -E arrow # Support for the CSV, Parquet, ORC and IPC/Feather/Arrow formats\npoetry install -E dgl # DGL support (also includes torch)\n```\n\nTo run the tests, make sure you have an [active Memgraph instance](/memgraph), and execute one of the following commands:\n\n```bash\npoetry run pytest . -k "not slow" # If all extras installed\n\npoetry run pytest . -k "not slow and not extras" # Otherwise\n```\n\nIf you’ve installed only certain extras, it’s also possible to run their associated tests:\n\n```bash\npoetry run pytest . -k "arrow"\npoetry run pytest . -k "dgl"\n```\n\n## GQLAlchemy capabilities\n\n<details>\n<summary>🗺️ Object graph mapper</summary>\n<br>\n\nBelow you can see an example of how to create `User` and `Language` node classes, and a relationship class of type `SPEAKS`. Along with that, you can see how to create a new node and relationship and how to save them in the database. After that, you can load those nodes and relationship from the database.\n<br>\n<br>\n\n```python\nfrom gqlalchemy import Memgraph, Node, Relationship, Field\nfrom typing import Optional\n\ndb = Memgraph()\n\nclass User(Node, index=True, db=db):\n    id: str = Field(index=True, exist=True, unique=True, db=db)\n\nclass Language(Node):\n    name: str = Field(unique=True, db=db)\n\nclass Speaks(Relationship, type="SPEAKS"):\n    pass\n\nuser = User(id="3", username="John").save(db)\nlanguage = Language(name="en").save(db)\nspeaks_rel = Speaks(\n    _start_node_id = user._id,\n    _end_node_id = language._id\n).save(db)\n\nloaded_user = User(id="3").load(db=db)\nprint(loaded_user)\nloaded_speaks = Speaks(\n        _start_node_id=user._id,\n        _end_node_id=language._id\n    ).load(db)\nprint(loaded_speaks)\n```\n</details>\n\n<details>\n<summary>🔨 Query builder</summary>\n<br>\nWhen building a Cypher query, you can use a set of methods that are wrappers around Cypher clauses. \n<br>\n<br>\n\n```python\nfrom gqlalchemy import create, match\nfrom gqlalchemy.query_builder import Operator\n\nquery_create = create()\n        .node(labels="Person", name="Leslie")\n        .to(relationship_type="FRIENDS_WITH")\n        .node(labels="Person", name="Ron")\n        .execute()\n\nquery_match = match()\n        .node(labels="Person", variable="p1")\n        .to()\n        .node(labels="Person", variable="p2")\n        .where(item="p1.name", operator=Operator.EQUAL, literal="Leslie")\n        .return_(results=["p1", ("p2", "second")])\n        .execute()\n```\n</details>\n\n<details>\n<summary>🚰 Manage streams</summary>\n<br>\n\nYou can create and start Kafka or Pulsar stream using GQLAlchemy. \n<br>\n\n**Kafka stream** \n```python\nfrom gqlalchemy import MemgraphKafkaStream\n\nstream = MemgraphKafkaStream(name="ratings_stream", topics=["ratings"], transform="movielens.rating", bootstrap_servers="localhost:9093")\ndb.create_stream(stream)\ndb.start_stream(stream)\n```\n\n\n**Pulsar stream**\n```python\nfrom gqlalchemy import MemgraphPulsarStream\n\nstream = MemgraphPulsarStream(name="ratings_stream", topics=["ratings"], transform="movielens.rating", service_url="localhost:6650")\ndb.create_stream(stream)\ndb.start_stream(stream)\n```\n\n</details>\n\n<details>\n<summary>🗄️ Import table data from different sources</summary>\n<br>\n\n**Import table data to a graph database**\n\nYou can translate table data from a file to graph data and import it to Memgraph. Currently, we support reading of CSV, Parquet, ORC and IPC/Feather/Arrow file formats via the PyArrow package.\n\nRead all about it in [table to graph importer how-to guide](https://memgraph.com/docs/gqlalchemy/how-to-guides/table-to-graph-importer).\n\n**Make a custom file system importer**\n\nIf you want to read from a file system not currently supported by GQLAlchemy, or use a file type currently not readable, you can implement your own by extending abstract classes `FileSystemHandler` and `DataLoader`, respectively.\n\nRead all about it in [custom file system importer how-to guide](https://memgraph.com/docs/gqlalchemy/how-to-guides/custom-file-system-importer).\n\n</details>\n\n<details>\n<summary>⚙️ Manage Memgraph instances</summary>\n<br>\n\nYou can start, stop, connect to and monitor Memgraph instances with GQLAlchemy.\n\n**Manage Memgraph Docker instance**\n\n```python\nfrom gqlalchemy.instance_runner import (\n    DockerImage,\n    MemgraphInstanceDocker\n)\n\nmemgraph_instance = MemgraphInstanceDocker(\n    docker_image=DockerImage.MEMGRAPH, docker_image_tag="latest", host="0.0.0.0", port=7687\n)\nmemgraph = memgraph_instance.start_and_connect(restart=False)\n\nmemgraph.execute_and_fetch("RETURN \'Memgraph is running\' AS result"))[0]["result"]\n```\n\n**Manage Memgraph binary instance**\n\n```python\nfrom gqlalchemy.instance_runner import MemgraphInstanceBinary\n\nmemgraph_instance = MemgraphInstanceBinary(\n    host="0.0.0.0", port=7698, binary_path="/usr/lib/memgraph/memgraph", user="memgraph"\n)\nmemgraph = memgraph_instance.start_and_connect(restart=False)\n\nmemgraph.execute_and_fetch("RETURN \'Memgraph is running\' AS result"))[0]["result"]\n```\n</details>\n\n<details>\n<summary>🔫 Manage database triggers</summary>\n<br>\n\nBecause Memgraph supports database triggers on `CREATE`, `UPDATE` and `DELETE` operations, GQLAlchemy also implements a simple interface for maintaining these triggers.\n\n```python\nfrom gqlalchemy import Memgraph, MemgraphTrigger\nfrom gqlalchemy.models import (\n    TriggerEventType,\n    TriggerEventObject,\n    TriggerExecutionPhase,\n)\n\ndb = Memgraph()\n\ntrigger = MemgraphTrigger(\n    name="ratings_trigger",\n    event_type=TriggerEventType.CREATE,\n    event_object=TriggerEventObject.NODE,\n    execution_phase=TriggerExecutionPhase.AFTER,\n    statement="UNWIND createdVertices AS node SET node.created_at = LocalDateTime()",\n)\n\ndb.create_trigger(trigger)\ntriggers = db.get_triggers()\nprint(triggers)\n```\n</details>\n\n<details>\n<summary>💽 On-disk storage</summary>\n<br>\n\nSince Memgraph is an in-memory graph database, the GQLAlchemy library provides an on-disk storage solution for large properties not used in graph algorithms. This is useful when nodes or relationships have metadata that doesn’t need to be used in any of the graph algorithms that need to be carried out in Memgraph, but can be fetched after. Learn all about it in the [on-disk storage how-to guide](https://memgraph.com/docs/gqlalchemy/how-to-guides/on-disk-storage).\n</details>\n\n<br>\n\nIf you want to learn more about OGM, query builder, managing streams, importing data from different source, managing Memgraph instances, managing database triggers and using on-disk storage, check out the GQLAlchemy [how-to guides](https://memgraph.com/docs/gqlalchemy/how-to-guides).\n\n## Development (how to build)\n\n```bash\npoetry run flake8 .\npoetry run black .\npoetry run pytest . -k "not slow and not extras"\n```\n\n## Documentation\n\nThe GQLAlchemy documentation is available on [memgraph.com/docs/gqlalchemy](https://memgraph.com/docs/gqlalchemy/).\n\nThe documentation can be generated by executing:\n```\npip3 install pydoc-markdown\npydoc-markdown\n```\n\n## License\n\nCopyright (c) 2016-2022 [Memgraph Ltd.](https://memgraph.com)\n\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use\nthis file except in compliance with the License. You may obtain a copy of the\nLicense at\n\n     http://www.apache.org/licenses/LICENSE-2.0\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.\n',
    'author': 'Bruno Sacaric',
    'author_email': 'bruno.sacaric@memgraph.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/memgraph/gqlalchemy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Copyright (c) 2016-2022 Memgraph Ltd. [https://memgraph.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import time

DATABASE_MISSING_IN_FIELD_ERROR_MESSAGE = """
Can't have an index on a property without providing the database `db` object.
Define your property as:
  {field}: {field_type} = Field({constraint}=True, db=Memgraph())
"""

DATABASE_MISSING_IN_NODE_CLASS_ERROR_MESSAGE = """
Can't have an index on a label without providing the database `db` object.
Define your class as:
  {cls.__name__}(Node, index=True, db=Memgraph())
"""

SUBCLASS_NOT_FOUND_WARNING = """
GraphObject subclass(es) '{types}' not found.
'{cls.__name__}' will be used until you create a subclass.
"""

ON_DISK_PROPERTY_DATABASE_NOT_DEFINED_ERROR = """
Error: Saving a node with an on_disk property without specifying an on disk database.

Add an on_disk_db like this:

from gqlalchemy import Memgraph, SQLitePropertyDatabase

db = Memgraph()
SQLitePropertyDatabase("path-to-sqlite-db", db)
"""

MISSING_OPTIONAL_DEPENDENCY = """
No module named '{dependency_name}'
"""

MISSING_ORDER = """
The second argument of the tuple must be order: ASC, ASCENDING, DESC or DESCENDING.
"""

ORDER_BY_TYPE_ERROR = """
TypeError: The argument provided is of wrong type. Please provide str, tuple[str, str] or list[tuple[str, str]].
"""

LITERAL_AND_EXPRESSION_MISSING = """
Can't create {clause} query without providing either 'literal' or 'expression' keyword arguments,
that can be literals, labels or properties.
"""

EXTRA_KEYWORD_ARGUMENTS = """
Can't create {clause} query with extra keyword arguments:
Please provide a value to either 'literal' or 'expression' keyword arguments.
"""

RESULT_QUERY_TYPE_ERROR = """
Can't create {clause} query:
The argument provided is of wrong type. Please provide str, tuple[str, str], list[Union[tuple[str, str], str]] or set[Union[tuple[str, str], str]].
"""

INSTANTIATION_ERROR = """
{class_name} class shouldn't be instantiatied!
"""

TOO_LARGE_TUPLE_IN_RESULT_QUERY = """
Tuple argument in {clause} clause only has two arguments - variable name and alias.
"""

FILE_NOT_FOUND = """
File with path {path} not found.
"""

OPERATOR_TYPE_ERROR = """
Operator argument in {clause} clause that is a string must be a valid operator.
"""

TIMEOUT_ERROR_MESSAGE = "Waited too long for the port {port} on host {host} to start accepting connections."
DOCKER_TIMEOUT_ERROR_MESSAGE = "Waited too long for the Docker container to start."
MEMGRAPH_CONNECTION_ERROR_MESSAGE = "The Memgraph process probably died."


class QueryClause(Enum):
    WHERE = "WHERE"
    SET = "SET"


class GQLAlchemyWarning(Warning):
    pass


class GQLAlchemySubclassNotFoundWarning(GQLAlchemyWarning):
    def __init__(self, types, cls):
        self.message = SUBCLASS_NOT_FOUND_WARNING.format(types=types, cls=cls)


class GQLAlchemyError(Exception):
    pass


class GQLAlchemyUniquenessConstraintError(GQLAlchemyError):
    pass


class GQLAlchemyDatabaseMissingInFieldError(GQLAlchemyError):
    def __init__(self, constraint: str, field: str, field_type: str):
        self.message = DATABASE_MISSING_IN_FIELD_ERROR_MESSAGE.format(
            constraint=constraint,
            field=field,
            field_type=field_type,
        )


class GQLAlchemyDatabaseMissingInNodeClassError(GQLAlchemyError):
    def __init__(self, cls):
        self.message = DATABASE_MISSING_IN_NODE_CLASS_ERROR_MESSAGE.format(cls=cls)


class GQLAlchemyOnDiskPropertyDatabaseNotDefinedError(GQLAlchemyError):
    def __init__(self):
        self.message = ON_DISK_PROPERTY_DATABASE_NOT_DEFINED_ERROR


class GQLAlchemyMissingOrder(GQLAlchemyError):
    def __init__(self):
        self.message = MISSING_ORDER


class GQLAlchemyOrderByTypeError(GQLAlchemyError):
    def __init__(self):
        self.message = ORDER_BY_TYPE_ERROR


class GQLAlchemyLiteralAndExpressionMissing(GQLAlchemyError):
    def __init__(self, clause: str):
        self.message = LITERAL_AND_EXPRESSION_MISSING.format(clause=clause)


class GQLAlchemyExtraKeywordArguments(GQLAlchemyError):
    def __init__(self, clause: str):
        self.message = EXTRA_KEYWORD_ARGUMENTS.format(clause=clause)


class GQLAlchemyTooLargeTupleInResultQuery(GQLAlchemyError):
    def __init__(self, clause) -> None:
        self.message = TOO_LARGE_TUPLE_IN_RESULT_QUERY.format(clause=clause)


class GQLAlchemyResultQueryTypeError(GQLAlchemyError):
    def __init__(self, clause):
        self.message = RESULT_QUERY_TYPE_ERROR.format(clause=clause)


class GQLAlchemyInstantiationError(GQLAlchemyError):
    def __init__(self, class_name) -> None:
        self.message = INSTANTIATION_ERROR.format(class_name=class_name)


class GQLAlchemyDatabaseError(GQLAlchemyError):
    def __init__(self, message):
        self.message = message


class GQLAlchemyOperatorTypeError(GQLAlchemyError):
    def __init__(self, clause) -> None:
        self.message = OPERATOR_TYPE_ERROR.format(clause=clause)


class GQLAlchemyTimeoutError(GQLAlchemyError):
    def __init__(self, message):
        self.message = message


class GQLAlchemyWaitForPortError(GQLAlchemyTimeoutError):
    def __init__(self, port, host):
        super().__init__(message=TIMEOUT_ERROR_MESSAGE.format(port=port, host=host))


class GQLAlchemyWaitForDockerError(GQLAlchemyTimeoutError):
    def __init__(self):
        super().__init__(message=DOCKER_TIMEOUT_ERROR_MESSAGE)


class GQLAlchemyWaitForConnectionError(GQLAlchemyTimeoutError):
    def __init__(self):
        super().__init__(message=MEMGRAPH_CONNECTION_ERROR_MESSAGE)


class GQLAlchemyFileNotFoundError(GQLAlchemyError):
    def __init__(self, path):
        super().__init__()
        self.message = FILE_NOT_FOUND.format(path=path)


def raise_if_not_imported(dependency, dependency_name):
    if not dependency:
        raise ModuleNotFoundError(MISSING_OPTIONAL_DEPENDENCY.format(dependency_name=dependency_name))


def database_error_handler(func):
    def inner_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise GQLAlchemyDatabaseError(e) from e

    return inner_function


def connection_handler(func, delay: float = 0.01, timeout: float = 5.0, backoff: int = 2):
    """Wrapper for a wait on the connection.

    Args:
        func: A function that tries to create the connection
        delay: A float that defines how long to wait between retries.
        timeout: A float that defines how long to wait for the port.
        backoff: An integer used for multiplying the delay.

    Raises:
      GQLAlchemyWaitForConnectionError: Raises an error
      after the timeout period has passed.
    """

    def _handler(*args, **kwargs):
        start_time = time.perf_counter()
        current_delay = delay
        while True:
            try:
                return func(*args, **kwargs)
            except Exception as ex:
                time.sleep(current_delay)
                if time.perf_counter() - start_time >= timeout:
                    raise GQLAlchemyWaitForConnectionError from ex

                current_delay *= backoff

    return _handler
